from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from concurrent.futures import as_completed
from concurrent.futures import ProcessPoolExecutor
from dataclasses import dataclass
from typing import Callable
from typing import cast
from typing import Iterator
from typing import Sequence
from typing import TypeVar

import pandas as pd

from .resourcefiles import MzMLResourceFile
from .utils import array_split
from .utils import PeptideSettings

Result = TypeVar("Result")


def get_processpool(workers: int) -> ProcessPoolExecutor:
    return ProcessPoolExecutor(max_workers=workers)


def parallel_result(
    exe: list[Callable[[], Result]],
    workers: int = 4,
) -> Iterator[Result]:
    if not exe:  # pragma: no cover
        return
    if workers <= 1:
        for e in exe:
            yield e()
        return
    with get_processpool(workers) as executor:
        futures = [executor.submit(e) for e in exe]
        for future in as_completed(futures):
            yield future.result()


def apply_func(series: pd.Series, func: Callable[..., Result]) -> pd.Series:
    return series.apply(func)  # type: ignore


def parallel_apply(
    series: pd.Series,
    func: Callable[..., Result],
    workers: int,
) -> pd.Series:  # pragma: no cover
    from functools import partial

    if workers <= 1:
        return apply_func(series, func)

    dfs: list[Callable[[], pd.Series]]

    dfs = [
        partial(apply_func, cast(pd.Series, sdf), func)
        for sdf in array_split(series, workers)
    ]

    return pd.concat(
        list(parallel_result(dfs, workers)),
        axis=0,
        ignore_index=False,
    )


class TaskBase(ABC):
    mzml: MzMLResourceFile

    @abstractmethod
    def task_run(self) -> pd.DataFrame: ...  # pragma: no cover


@dataclass
class TaskData:
    pepxml_df: pd.DataFrame
    mzml: MzMLResourceFile
    settings: PeptideSettings
    level: int = 0
    number_of_bg_processes: int = 1

    def mem(self) -> int:  # pragma: no cover
        return self.pepxml_df.memory_usage(deep=True).sum()


class Task(TaskData, TaskBase):
    pass


def parallel_tasks(
    tasks: Sequence[TaskBase],
    *,
    workers: int = 4,
) -> Iterator[tuple[pd.DataFrame, MzMLResourceFile]]:
    if not tasks:
        return

    if workers <= 1 or len(tasks) == 1:
        for task in tasks:
            yield task.task_run(), task.mzml
        return
    # ProcessPoolExecutor re-intialize loggers in the worker processes
    # with initialize function?
    nworkers = min(len(tasks), workers)

    with get_processpool(nworkers) as executor:
        futures = {executor.submit(t.task_run): t for t in tasks}
        for future in as_completed(futures):
            yield future.result(), futures[future].mzml



from __future__ import annotations

import re
from pathlib import Path

import numpy as np
import pandas as pd
from sqlalchemy import MetaData
from sqlalchemy import select
from sqlalchemy import text
from sqlalchemy import update
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker

from .model import Peptide
from .utils import file2engine


def get_pepitde_cols(engine: Engine) -> list[str]:
    """Get *actual* list of column names for peptides table."""
    assert engine.url.drivername == "sqlite", engine.url
    q = text(f"PRAGMA table_info({Peptide.__tablename__})")
    with engine.connect() as conn:
        tcols = conn.execute(q).all()
        return [row[1] for row in tcols]


def get_pepitde_cols2(engine: Engine) -> list[str]:
    """Get *actual* list of column names for peptides table."""
    m = MetaData()
    m.reflect(bind=engine, only=[Peptide.__tablename__])
    pep = m.tables[Peptide.__tablename__]
    return [c.name for c in pep.columns]


def add_pep_columns(engine: Engine, cols: list[str]) -> None:
    """Add columns to peptides table."""
    tablename = Peptide.__tablename__
    q = "ALTER TABLE {tablename} add {col} INTEGER NOT NULL DEFAULT 0"
    with engine.connect() as conn:
        for col in cols:
            conn.execute(text(q.format(col=col, tablename=tablename)))
        conn.commit()


def drop_pep_columns(engine: Engine, cols: list[str]) -> None:
    """Drop columns to peptides table."""
    tablename = Peptide.__tablename__
    q = "ALTER TABLE {tablename} DROP COLUMN {col}"
    with engine.connect() as conn:
        for col in cols:
            conn.execute(text(q.format(col=col, tablename=tablename)))
        conn.commit()


def addif_pep_columns(engine: Engine, cols: list[str]) -> list[str]:
    """Add columns that don't exist in peptides table."""
    rcols = get_pepitde_cols(engine)
    cols = list(set(cols) - set(rcols))
    if cols:
        add_pep_columns(engine, cols)
    return cols


def dropif_pep_columns(
    engine: Engine,
    cols: list[str],
    *,
    regex: bool = False,
) -> list[str]:
    """Remove columns that exist in peptides table."""
    rcols = get_pepitde_cols(engine)
    if regex:
        xcols = set()
        for cc in rcols:
            for c in cols:
                if re.match(c, cc) is not None:
                    xcols.add(cc)
                    break
        cols = list(xcols)
    else:
        cols = list(set(cols) - set(rcols))
    if cols:
        drop_pep_columns(engine, cols)
    return cols


def isoPeakNr(arr: list[float] | None, isoPeaksMinRSQ: float) -> int:
    if not arr:
        return 0
    return int(np.sum(np.array(arr[2:]) >= isoPeaksMinRSQ))  # need this int()!


def populate_isoPeaksNr(
    engine: Engine,
    cols: list[tuple[str, float]],
    force: bool = False,
) -> list[str]:
    rcols = addif_pep_columns(engine, [c[0] for c in cols])
    if not rcols and not force:
        return rcols
    q = select(Peptide.peptideid, Peptide.adjustedRsq)
    with engine.connect() as conn:
        res = conn.execute(q).all()

    Session = sessionmaker()
    # see "ORM Bulk UPDATE by Primary Key"
    # https://docs.sqlalchemy.org/en/20/orm/queryguide/dml.html#orm-bulk-update-by-primary-key

    with Session(bind=engine) as session:
        for col, isoPeaksMinRSQ in cols:
            u = [
                {"peptideid": peptideid, col: isoPeakNr(adjustedRsq, isoPeaksMinRSQ)}
                for (peptideid, adjustedRsq) in res
            ]
            session.execute(update(Peptide), u)

        session.commit()

    return rcols


def isoPeaksMinRSQToI(r2: float) -> int:
    return int(r2 * 100)


def gen_cols() -> list[tuple[str, float]]:
    from .model import MINRSQ100

    return [(f"iso_peaks_nr_{n}", n / 100.0) for n in MINRSQ100]


def populate_isoPeaksNrFixed(
    engine: Engine,
    force: bool = False,
) -> list[str]:
    return populate_isoPeaksNr(
        engine,
        gen_cols(),
        force=force,
    )


## @export
def fixsqlite(
    filename: str | Path,
    *,
    force: bool = False,
    echo: bool = False,
) -> list[str]:
    path = Path(filename)
    if not path.exists():
        return []
    engine = file2engine(filename, echo=echo)
    try:
        return populate_isoPeaksNrFixed(engine, force=force)
    finally:
        engine.dispose()


def to_nr(arr: list[float], cols: list[tuple[str, float]]) -> pd.Series:
    if not arr:
        return pd.Series({col: 0 for col, _ in cols})
    a = np.array(arr)
    return pd.Series(
        {
            col: int(np.sum(np.array(a[2:]) >= isoPeaksMinRSQ))
            for col, isoPeaksMinRSQ in cols
        },
    )


def add_iso_peaks(pep_df: pd.DataFrame) -> pd.DataFrame:
    rsq = pep_df["adjustedRsq"]

    edf = rsq.apply(to_nr, args=(gen_cols(),))
    return pd.concat([pep_df, edf], axis=1)



from __future__ import annotations

import os
import signal
import subprocess
import sys
import time
import tomllib
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from shutil import rmtree
from tempfile import TemporaryDirectory
from threading import Thread
from types import ModuleType
from typing import Any
from typing import Callable
from typing import Literal
from typing import TypedDict

import click
import tomli_w
from typing_extensions import NotRequired
from typing_extensions import override

from .background import Processing
from .cli import CONFIG
from .utils import get_package
from .utils import has_package


class PopenArgs(TypedDict):
    process_group: NotRequired[int | None]
    creationflags: int
    preexec_fn: Callable[[], None] | None


@dataclass
class Runner:
    name: str
    cmd: list[str]
    directory: str = "."
    env: dict[str, str] | None = None
    showcmd: bool = False
    shell: bool = False
    prevent_sig: bool = False  # prevent Cntrl-C from propagating to child process

    def getenv(self) -> dict[str, str] | None:
        if not self.env:
            return None
        return {**os.environ, **self.env}

    def start(self) -> subprocess.Popen[bytes]:
        if self.showcmd:
            click.secho(" ".join(str(s) for s in self.cmd), fg="magenta")

        kwargs = PopenArgs(creationflags=0, preexec_fn=None)
        if self.prevent_sig:
            self.set_new_process_group(kwargs)
        return subprocess.Popen(  # type: ignore
            self.cmd,
            cwd=self.directory,
            env=self.getenv(),
            shell=self.shell,
            **kwargs,
        )

    def set_new_process_group(self, kwargs: PopenArgs) -> None:
        if sys.platform == "win32":
            kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP
        else:
            if sys.version_info >= (3, 11):
                kwargs["process_group"] = 0
            else:
                kwargs["preexec_fn"] = os.setpgrp


def wait_for_connection(url: str, wait: float = 10.0) -> bool:
    from urllib.request import Request, urlopen
    from urllib.error import URLError

    req = Request(url, method="HEAD")

    while wait > 0:
        time.sleep(1)
        wait -= 1
        try:
            urlopen(req)
            return True
        except URLError:
            pass
    return False


def browser(
    url: str = "http://127.0.0.1:8000",
    sleep: float = 10.0,
) -> Thread:
    import webbrowser

    def run() -> None:
        try:
            wait_for_connection(url, sleep)
        finally:
            webbrowser.open_new_tab(url)

    tr = Thread(target=run)
    tr.daemon = True  # exit when main process exits
    tr.start()
    return tr


def default_conf() -> dict[str, Any]:
    # from tempfile import gettempdir

    conf = {
        "MOUNTPOINTS": [
            ("~", "HOME"),
        ],
        "JOBSDIR": ["~/turnover_jobs"],
        "CACHEDIR": "~/turnover_cache",
        "WEBSITE_STATE": "single_user",
    }
    return conf


def instance_conf(config: Path | str, ns: str | None = None) -> dict[str, Any]:
    # """We *must* have flask in our environment by now"""
    # from flask import Config  # pylint: disable=import-error
    # conf = Config(".")
    # conf.from_pyfile(config)
    # return conf
    try:
        with open(config, "rb") as fp:
            ret = tomllib.load(fp)
            if ns is not None:
                ret2 = ret.get(ns, None)
                return ret2 or {}
            return ret
    except tomllib.TOMLDecodeError as e:  # pragma: no cover
        msg = f'Can\'t read configuration file "{config}": {e}'
        raise click.BadParameter(msg) from e


def dump_config(configfile: Path, with_website: bool) -> None:
    from . import config

    def mod2dict(module: ModuleType) -> dict[str, Any]:
        d: dict[str, Any] = {}
        for k in dir(module):
            if k.isupper():
                a = getattr(module, k)
                if a is None:
                    # a = ""
                    continue
                d[k] = a

        d = dict(sorted(d.items()))
        return d

    HEADER = b"""# turnover configuration file.
# Remove/comment out any values you don't change since these are defaults.
"""
    d = mod2dict(config)
    if with_website:
        c = get_package("protein_turnover_website.config")
        assert c is not None
        d["website"] = mod2dict(c)

    click.secho(f"writing: {configfile}")
    with open(configfile, "wb") as fp:
        fp.write(HEADER)
        tomli_w.dump(d, fp)


def waitress_app(wsgi_app: str, port: int) -> list[str]:
    return [
        sys.executable,
        "-m",
        "waitress",
        "--threads=6",
        f"--listen=127.0.0.1:{port}",
        wsgi_app + ":application",
    ]


def flask_app(wsgi_app: str, port: int) -> list[str]:
    return [sys.executable, "-m", "flask", "--app", wsgi_app, "run", f"--port={port}"]


def gunicorn_app(wsgi_app: str, port: int) -> list[str]:
    return [
        sys.executable,
        "-m",
        "gunicorn",
        f"--bind=127.0.0.1:{port}",
        wsgi_app,
    ]


APPS = {"gunicorn": gunicorn_app, "waitress": waitress_app, "flask": flask_app}

WebServer = Literal["flask", "gunicorn", "waitress"]


@dataclass(kw_only=True)
class WebRunner:
    browse: bool = False
    workers: int = 1
    web_config: str | Path | None = None  # from --web-config
    server: WebServer | None = None
    view_only: bool = True
    configfile: str | Path | None = None  # turnover config file
    defaults: dict[str, Any] | None = None  # CACHEDIR etc. from commandline
    port: int = 8000
    server_options: tuple[str, ...] = ()  # extra commandline arguments after --
    wsgi_app: str = "protein_turnover_website.wsgi"
    page: str | None = None
    mailhost: str | None = None
    sleep: float = 30.0
    prevent_sig: bool = True  # sys.platform != "win32"  # False on windows
    compress_result: bool = False  # gzip results files

    def get_server(self) -> WebServer:  # pragma: no cover
        if not has_package("protein_turnover_website"):
            click.secho(
                "Please install protein_turnover_website [pip install protein-turnover-website]!",
                fg="red",
                err=True,
            )
            raise click.Abort()
        server = self.server
        if server is None:  # try for a better server
            s: WebServer
            for s in ["gunicorn", "waitress", "flask"]:  # type: ignore
                if has_package(s):
                    server = s
                    break
        else:
            if not has_package(server):
                click.secho(
                    f"Please install {self.server} [pip install {self.server}]!",
                    fg="red",
                    err=True,
                )
                raise click.Abort()
        assert server is not None
        return server

    def get_webconf(self) -> dict[str, Any]:
        web_config = self.web_config
        if web_config is None:
            if self.configfile is not None and Path(self.configfile).exists():
                web_config = self.configfile
            elif CONFIG.exists():
                web_config = CONFIG

        web_conf = default_conf()

        if web_config:
            # just need JOBSDIR
            ns = "website" if self.web_config is None else None

            click.secho(
                f"reading  web configuration from: {web_config} namespace={ns if ns else ''}",
                fg="green",
            )
            web_conf.update(
                instance_conf(web_config, ns=ns),
            )
        # from commandline so... last
        if self.defaults is not None:
            web_conf.update(self.defaults)

        if self.mailhost is not None:
            if self.mailhost == "none":
                web_conf["WANT_EMAIL"] = 0  # no point user specifying an email

        if self.compress_result:
            web_conf["COMPRESS_RESULT"] = True

        # we don't want an email logger for a single use server
        if "LOG_MAIL_SERVER" not in web_conf or self.mailhost == "none":
            web_conf["LOG_MAIL_SERVER"] = "none"
        return web_conf

    def get_website(self, server: WebServer, jobfile: Path) -> Runner:
        return Runner(
            server,
            [
                *APPS[server](self.wsgi_app, self.port),
                *self.server_options,
            ],
            env={"TURNOVER_SETTINGS": str(jobfile)},
            prevent_sig=self.prevent_sig,
        )

    def get_mailhost(self) -> str | None:
        return self.mailhost
        # from .config import MAIL_SERVER

        # if self.mailhost is not None:
        #     return self.mailhost
        # return MAIL_SERVER

    def cleanup(self, jobsdir: Path) -> None:
        # under windows we user terminate so the
        # background process will leave the pid file around maybe
        from .background import PID_FILENAME

        pidfile = jobsdir / PID_FILENAME
        pidfile.unlink(missing_ok=True)
        # if not sys.platform == 'win32':
        #     return
        torm = set()
        for d, _, files in jobsdir.walk():
            for f in files:
                if f.endswith(".toml.pid"):
                    torm.add(d)
                    break
        for d in torm:
            click.secho(f"removing partial calculation: {d}", fg="yellow")
            rmtree(d, ignore_errors=True)

    def run(self) -> bool:
        """Run full website."""

        server = self.get_server()
        web_conf = self.get_webconf()

        # need to read config file just for jobsdir
        jd = web_conf["JOBSDIR"]
        if isinstance(jd, list):
            jd = jd[0]
        jobsdir = Path(jd).expanduser()
        if not jobsdir.exists():
            jobsdir.mkdir(parents=True, exist_ok=True)
        background = None
        if not self.view_only:
            m = self.get_mailhost()
            email_args = [] if m is None else [f"--mailhost={m}"]
            cfg = [f"--config={self.configfile}"] if self.configfile is not None else []
            if self.compress_result:
                email_args.append("--compress-result")
            background = Runner(
                "background",
                [
                    sys.executable,
                    "-m",
                    "protein_turnover",
                    *cfg,
                    "--level=info",
                    "background",
                    f"--workers={self.workers}",
                    *email_args,
                    str(jobsdir),
                ],
                directory=".",
                prevent_sig=self.prevent_sig,
            )
        Url = f"127.0.0.1:{self.port}"
        if self.page is not None:
            Url += f"/{self.page}"

        # ON windows NamedTemporaryFile can't be read
        # by other processes, so we use a directory
        # ... which seems to work... sigh!
        with TemporaryDirectory() as td:
            filename = Path(td).absolute() / "turnover-web.toml"
            with filename.open("wb") as fp:
                tomli_w.dump(web_conf, fp)
            assert filename.exists()

            website = self.get_website(server, filename)

            if self.view_only:
                procs = [website]
            else:
                assert background is not None
                procs = [background, website]

            processes = [(p.name, p.start()) for p in procs]

            if self.browse:
                browser(url=f"http://{Url}", sleep=10.0)
            return self.loop(jobsdir, processes)

    def loop(
        self,
        jobsdir: Path,
        processes: list[tuple[str, subprocess.Popen[bytes]]],
    ) -> bool:
        worker = Processing(jobsdir)
        ninterrupts = 0
        prev = datetime.now()
        while True:
            try:
                time.sleep(self.sleep)
                failed = False
                done = 0
                for n, tr in processes:
                    try:
                        retcode = tr.wait(1.0)
                        if retcode != 0:
                            click.echo(f"process failed {n}")
                            failed = True
                        else:
                            done += 1
                    except subprocess.TimeoutExpired:
                        pass
                if failed:
                    raise KeyboardInterrupt()
                if done == len(processes):  # only in testing...
                    return True

            except KeyboardInterrupt:
                # too long between ^C
                now = datetime.now()
                if (
                    ninterrupts > 0
                    and not self.view_only
                    and (now - prev).total_seconds() > 5
                ):
                    ninterrupts = 0
                    prev = now
                    continue
                ninterrupts += 1
                if ninterrupts >= 2 or self.view_only:
                    for name, tr in processes:
                        click.secho(f"terminating... {name}", fg="magenta")
                        if sys.platform == "win32":
                            tr.terminate()
                        else:
                            tr.send_signal(signal.SIGINT)
                    for name, tr in processes:
                        try:
                            tr.wait(timeout=4.0)
                        except (OSError, subprocess.TimeoutExpired):
                            pass
                    self.cleanup(jobsdir)
                    sys.exit(os.EX_OK)

                prev = now

                if not self.view_only and worker.is_processing():
                    click.secho(
                        "Warning! The background process is running a job!",
                        fg="yellow",
                        bold=True,
                    )
                click.secho("interrupt... ^C again to terminate")
        return False


class TestWebRunner(WebRunner):
    @override
    def get_server(self) -> WebServer:
        return "flask"

    @override
    def get_website(self, server: WebServer, jobfile: Path) -> Runner:
        return Runner(
            "test-website",
            [
                sys.executable,
                "-m",
                "protein_turnover.web",  # see "test website below"
                str(jobfile),
            ],
            directory=".",
        )


if __name__ == "__main__":
    # used in testing....
    import click

    @click.command()
    @click.argument("jobfile")
    def testwebsite(jobfile: str):
        """Minimal website"""
        tomlfile = Path(jobfile)
        with tomlfile.open("rb") as fp:
            tomllib.load(fp)

        time.sleep(0.5)  # do some website stuff

    testwebsite()



from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from typing import Callable
from typing import TYPE_CHECKING

from . import current_fit
from . import dec2023_fit

if TYPE_CHECKING:
    from .utils import PeptideSettings
    from .utils import PeptideInfo
    import numpy as np
    from .types.numpytypes import NP1DF64Array, NP2DF64Array, NP1DF32Array


class Algo(ABC):
    @staticmethod
    @abstractmethod
    def mk_maxIso(settings: PeptideSettings) -> Callable[[str], np.int32]: ...

    # @staticmethod
    # @abstractmethod
    # def isotopic_distribution(
    #     pepinfo: PeptideInfo, abundance: float | None = None
    # ) -> NP1DF64Array: ...

    @staticmethod
    @abstractmethod
    def make_envelope_array(
        pepinfo: PeptideInfo,
        maxIso: int,
    ) -> tuple[NP1DF64Array, NP2DF64Array]: ...
    @staticmethod
    @abstractmethod
    def natural_dist(
        pepinfo: PeptideInfo,
        monoScale: float,
    ) -> NP1DF32Array: ...


class AlgoLatest(Algo):
    @staticmethod
    def mk_maxIso(settings: current_fit.PeptideSettings) -> Callable[[str], np.int32]:
        return current_fit.mk_maxIso(settings)

    # @staticmethod
    # def isotopic_distribution(
    #     pepinfo: PeptideInfo, abundance: float | None = None
    # ) -> NP1DF64Array:
    #     return current_fit.isotopic_distribution(pepinfo, abundance)

    @staticmethod
    def make_envelope_array(
        pepinfo: PeptideInfo,
        maxIso: int,
    ) -> tuple[NP1DF64Array, NP2DF64Array]:
        return current_fit.make_envelope_array(pepinfo, maxIso)

    @staticmethod
    def natural_dist(
        pepinfo: PeptideInfo,
        monoScale: float,
    ) -> NP1DF32Array:
        return current_fit.natural_dist(pepinfo, monoScale)


class Algo2023(Algo):
    @staticmethod
    def mk_maxIso(settings: current_fit.PeptideSettings) -> Callable[[str], np.int32]:
        return dec2023_fit.mk_maxIso(settings)

    # @staticmethod
    # def isotopic_distribution(
    #     pepinfo: PeptideInfo, abundance: float | None = None
    # ) -> NP1DF64Array:
    #     return dec2023_fit.isotopic_distribution(pepinfo, abundance)

    @staticmethod
    def make_envelope_array(
        pepinfo: PeptideInfo,
        maxIso: int,
    ) -> tuple[NP1DF64Array, NP2DF64Array]:
        return dec2023_fit.make_envelope_array(pepinfo, maxIso)

    @staticmethod
    def natural_dist(
        pepinfo: PeptideInfo,
        monoScale: float,
    ) -> NP1DF32Array:
        return dec2023_fit.natural_dist(pepinfo, monoScale)


ALGO: dict[str, type[Algo]] = {
    "Latest": AlgoLatest,
    "2023": Algo2023,
}


def settings_to_algo(settings: PeptideSettings) -> type[Algo]:
    return ALGO[settings.pipelineAlgo]



from __future__ import annotations

import sys

import click

from .cli import cli
from .cli import Config
from .cli import pass_config


@cli.command(name="background")
@click.option(
    "--workers",
    type=int,
    help="number of workers [default - half the number of cpus]",
)
@click.option(
    "--nice",
    default=0,
    type=click.IntRange(
        min=0,
    ),
    help="run turnover command at this nice level",
)
@click.option(
    "--sleep",
    default=10.0
    if sys.platform == "win32"
    else 60.0,  # no signal for windows so shorter cycle
    help="sleep (seconds) between directory scans",
    show_default=True,
)
@click.option(
    "--run-config",
    type=click.Path(file_okay=True, dir_okay=False, exists=True),
    help="configuration file to pass off to the background process",
)
@click.option(
    "--compress-result",
    is_flag=True,
    help="compress the resultant sqlite file to save space",
)
@click.option(
    "--mailhost",
    help='send emails to this mailhost. use "none" for no emails',
    hidden=False,
)  # see runner.py
@click.argument(
    "jobs_directory",
    type=click.Path(dir_okay=True, exists=True, file_okay=False),
    required=True,
    nargs=-1,
)
@pass_config
def background_cmd(
    cfg: Config,
    jobs_directory: tuple[str, ...],
    workers: int | None,
    run_config: str | None,
    sleep: float,
    nice: int,
    mailhost: str | None = None,
    compress_result: bool = False,
) -> None:  # pragma: no cover
    """Watch and run turnover jobs from directory"""
    from os import cpu_count
    from .background import SimpleQueue
    from .logger import logger

    if workers is None:
        workers = max((cpu_count() or 1) // 2, 1)
        logger.warning("using %d workers", workers)

    squeue = SimpleQueue(
        jobs_directory,
        workers=workers,
        wait=sleep,
        nice=nice,
        config=run_config or cfg.user_config,
        mailhost=mailhost,
        compress_result=compress_result,
    )
    squeue.run()



from __future__ import annotations

import operator
from collections.abc import Sequence
from dataclasses import dataclass
from dataclasses import fields
from dataclasses import replace
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Literal
from typing import NamedTuple
from typing import NotRequired
from typing import TypedDict

import numpy as np
import pandas as pd  # type: ignore
from sqlalchemy import and_
from sqlalchemy import func
from sqlalchemy import join
from sqlalchemy import or_
from sqlalchemy import Select
from sqlalchemy import select
from sqlalchemy import Subquery
from sqlalchemy import text
from sqlalchemy.engine import Connection
from sqlalchemy.engine import Engine
from sqlalchemy.exc import OperationalError
from sqlalchemy.orm import DeclarativeBase
from sqlalchemy.orm import InstrumentedAttribute
from sqlalchemy.orm import Relationship
from sqlalchemy.sql import ColumnElement
from sqlalchemy.sql.expression import BinaryExpression
from typing_extensions import override

from .model import PepProt
from .model import Peptide
from .model import Protein
from .utils import file2engine
from .utils import precision as to_precision
from .utils import rehydrate_peptides

# from sqlalchemy.orm import load_only
# from sqlalchemy.orm import selectinload
# from sqlalchemy.orm import Session


# func.count()
# pylint: disable=not-callable
class WhereClauses(NamedTuple):
    pep_where: ColumnElement[bool] | None
    prot_where: ColumnElement[bool] | None


INT_FORM_FIELDS = {"isoPeaksMinRSQ", "sqlite_version", "isoPeaksNr", "sqlite_version"}


## @export
@dataclass
class FormFilter:
    peptideQualityFilter: float | None = None
    heavyMinCor: float | None = None
    isoPeaksNr: int | None = None
    isoPeaksMinRSQ: int = -1
    #    heavyMinRSQ: float | None = None
    monoMinus1MinRatio: float | None = None
    # enrichmentFactor: float | None = None
    enrichmentMax: float | None = None
    enrichmentMin: float | None = None
    fdrMaximum: float | None = None
    nnlsResidual: float | None = None
    maxPeakArea: float | None = None
    log_sigma_lpf: float | None = None
    sqlite_version: int = -1

    def calc_where(self, scale_intensity: bool = False) -> WhereClauses:
        pep_filters = []
        prot_filters = []

        def add(q: Any) -> None:
            pep_filters.append(q)

        if self.peptideQualityFilter is not None:
            add(Peptide.peptideprophet_probability >= self.peptideQualityFilter)

        if self.fdrMaximum is not None:
            add(Peptide.fdr <= self.fdrMaximum)

        if self.heavyMinCor is not None:
            add(Peptide.heavyCor >= self.heavyMinCor)

        # if self.heavyMinRSQ is not None:
        #     add(Peptide.heavy_adj_r_squared >= self.heavyMinRSQ)

        if self.nnlsResidual is not None:
            if scale_intensity:  # pragma: no cover
                add(
                    Peptide.nnls_residual / Peptide.totalIntensityWeight
                    <= self.nnlsResidual,
                )
            else:
                add(
                    Peptide.nnls_residual / Peptide.totalNNLSWeight
                    <= self.nnlsResidual,
                )

            # add(func.power(Peptide.nnls_residual,2) <= self.nnlsResidual)
            # add(Peptide.nnls_residual <= self.nnlsResidual)

        if self.monoMinus1MinRatio is not None and self.monoMinus1MinRatio > 0.0:
            add(Peptide.inv_ratio < 1.0 / self.monoMinus1MinRatio)

        if self.enrichmentMax is not None:
            add(Peptide.enrichment <= self.enrichmentMax)
        if self.enrichmentMin is not None:
            add(Peptide.enrichment >= self.enrichmentMin)

        if self.maxPeakArea is not None:
            add(Peptide.maxPeakArea >= self.maxPeakArea)

        if self.log_sigma_lpf is not None:
            prot_filters.append(
                func.log((Protein.lpf_std + 1e-4) / Protein.lpf_median)
                <= self.log_sigma_lpf,
            )

        if (
            self.sqlite_version >= 1
            and self.isoPeaksNr is not None
            and self.isoPeaksMinRSQ is not None
            and self.isoPeaksMinRSQ != -1  # not set
        ):
            col = getattr(Peptide, f"iso_peaks_nr_{int(self.isoPeaksMinRSQ)}", None)
            if col is not None:
                add(col >= int(self.isoPeaksNr))

        return WhereClauses(
            and_(*pep_filters) if pep_filters else None,
            and_(*prot_filters) if prot_filters else None,
        )


FILTER_NAMES = {f.name for f in fields(FormFilter)}


def ident(s: str) -> str:
    return s


## @export
def dict_to_form_filter(
    d: dict[str, str],
    match: Callable[[str], str | None] = ident,
) -> FormFilter:
    ret: dict[str, float | int] = {}

    for k in d.keys():
        col = match(k)
        if col is not None:
            if col in FILTER_NAMES:
                value = d[k]
                if value:
                    ret[col] = int(value) if col in INT_FORM_FIELDS else float(value)
    return FormFilter(**ret)  # type: ignore


## @export
def all_plotting_columns() -> list[Column]:  # pragma: no cover
    from ..plotting import ALL_COLUMNS

    return [Column(c) for c in ALL_COLUMNS]


def like(attr: InstrumentedAttribute, value: str) -> BinaryExpression[bool]:
    if "%" not in value:
        value = "%" + value + "%"
    return attr.like(value)


def in_(attr: InstrumentedAttribute, value: str) -> BinaryExpression[bool]:
    return attr.in_(value)


def doregex(attr: InstrumentedAttribute, value: str) -> BinaryExpression[bool]:
    return attr.op("REGEXP")(value)  # need pcre installed in sqllite


OPMAP: dict[str, Callable[..., Any]] = {
    ">": operator.gt,
    ">=": operator.ge,
    "=": operator.eq,
    "<": operator.lt,
    "<=": operator.le,
    "!=": operator.ne,
    "like": like,
    "regex": doregex,
    "in": in_,
}


## @export
class RowQuery(NamedTuple):
    column: str
    op: str
    value: Any

    def to_sql(self, Orm: type[DeclarativeBase]) -> ColumnElement[bool]:
        attr = getattr(Orm, self.column, None)
        if attr is None or not hasattr(attr, "property"):
            raise ValueError(
                f"unknown column {self.column} for {Orm.__tablename__} table",
            )
        if isinstance(attr.property, Relationship):  # pragma: no cover
            raise ValueError(
                f"column {self.column} is a relationship on {Orm.__tablename__}!",
            )
        op = OPMAP.get(self.op)
        if op is None:
            raise ValueError(f"no operator like {self.op}")

        return op(attr, self.value)


## @export
@dataclass
class RowFilter:
    rows: list[RowQuery]
    method: Literal["and", "or"] = "and"

    # used in protein_turnover_website
    def add(self, more: RowFilter) -> RowFilter:
        return replace(self, rows=self.rows + more.rows)

    def to_and_sql(self, Orm: type[DeclarativeBase]) -> ColumnElement[bool]:
        return and_(*[q.to_sql(Orm) for q in self.rows])

    def to_or_sql(
        self,
        Orm: type[DeclarativeBase],
    ) -> ColumnElement[bool]:  # pragma: no cover
        return or_(*[q.to_sql(Orm) for q in self.rows])

    def to_sql(self, Orm: type[DeclarativeBase]) -> ColumnElement[bool]:
        if self.method == "and":
            return self.to_and_sql(Orm)
        return self.to_or_sql(Orm)  # pragma: no cover


## @export
@dataclass
class DTQuery:
    start: int = 0
    length: int = 10
    search: str | None = None
    regex: bool = False
    ascending: bool = True
    order_column: str | None = None
    draw: int = 0  # data tables


def search_protein_tosql(search: str, search_columns: list[str]) -> ColumnElement[bool]:
    return RowFilter(
        [RowQuery(coln, "like", search) for coln in search_columns],
    ).to_or_sql(Protein)


## @export
@dataclass
class ProteinQueryResult:
    result_df: pd.DataFrame
    total_proteins: int = -1
    total_peptides: int = -1
    total_filtered: int = -1


def np_agg(val: str | None, func: Callable[[np.ndarray], float]) -> float | None:
    if val is None:
        return None
    values = np.array([float(v) for v in val.split(",")])
    if len(values) == 0:  # pragma: no cover
        return np.nan
    return func(values)


def std(val: str | None) -> float | None:
    return np_agg(val, np.std)


def median(val: str | None) -> float | None:
    return np_agg(val, np.median)


# sqlite is missing some aggregate functions... we "fake" them here.
FAKE_AGGREGATE_FUNCS = {"std": std, "median": median}


class AGDict(TypedDict):
    column: str
    aggfunc: str
    label: NotRequired[str]
    args: NotRequired[Sequence[str] | str]
    denom: NotRequired[str]


# see https://www.sqlite.org/lang_aggfunc.html
## @export
@dataclass(kw_only=True)
class Aggregate:
    column: str
    aggfunc: str  # aggregate function sum,avg,group_concat etc.
    label: str = ""
    args: Sequence[str] | str = ()

    def __post_init__(self) -> None:
        if self.label == "":
            self.label = self.column
        if isinstance(self.args, str):  # pragma: no cover
            self.args = [self.args]

    def getattr(
        self,
        orm: type[DeclarativeBase],
        column: str,
    ) -> ColumnElement[Any]:
        attr = getattr(orm, column, None)
        if attr is None or not hasattr(attr, "property"):
            raise ValueError(f"unknown column {column} for {orm.__tablename__} table")
        if isinstance(attr.property, Relationship):  # pragma: no cover
            raise ValueError(
                f"column {column} is a relationship on {orm.__tablename__}!",
            )
        return attr

    def getcolumn(self, orm: type[DeclarativeBase]) -> ColumnElement[Any]:
        return self.getattr(orm, self.column)

    def make_column(self, orm: type[DeclarativeBase]) -> ColumnElement[Any]:
        attr = self.getcolumn(orm)

        if self.aggfunc in FAKE_AGGREGATE_FUNCS:
            aggf = getattr(func, "group_concat")
            attr = aggf(attr, ",")
        else:
            aggf = getattr(func, self.aggfunc)
            attr = aggf(attr, *self.args)
        return attr.label(self.label)


## @export
@dataclass(kw_only=True)
class RatioAggregate(Aggregate):
    denom: str = "totalNNLSWeight"

    @override
    def getcolumn(self, orm: type[DeclarativeBase]) -> ColumnElement[Any]:
        attr = self.getattr(orm, self.column)
        denom = self.getattr(orm, self.denom)
        return attr / denom


## @export
@dataclass
class Column:
    name: str
    title: str = ""  # Title for datatables.net
    read: bool = True  # read this column from data table
    send: bool = True  # send this column to browser
    view: bool = True  # show this column as datatables column
    search: bool = True  # search this column given search query
    aggregate: Aggregate | None = None  # is an aggregate column
    className: str | None = None  # possible className for datatables.net
    denom: str | None = None
    label: str = ""

    def __post_init__(self) -> None:
        if self.title == "":
            self.title = self.name
        if self.label == "":
            self.label = self.name
        if self.aggregate is not None:
            if not isinstance(self.aggregate, Aggregate):  # pragma: no cover
                if "denom" in self.aggregate:
                    self.aggregate = RatioAggregate(**self.aggregate)
                else:
                    self.aggregate = Aggregate(**self.aggregate)
            if self.aggregate.label != self.name:
                self.aggregate = replace(self.aggregate, label=self.name)
            self.search = False
            self.read = False

    def getcolumn(self, orm: type[DeclarativeBase]) -> ColumnElement[Any] | None:
        ret = getattr(orm, self.name, None)
        if ret is None:
            return None
        if self.denom is not None:
            denom = getattr(orm, self.denom, None)
            if denom is None:
                return None
            return (ret / denom).label(self.label)
        if self.label != self.name:
            ret = ret.label(self.label)
        return ret


## @export
class ProteinQuery:
    Join = join(PepProt, Peptide)
    QID = select(PepProt.proteinid).select_from(Join)
    NPRO = select(func.count(PepProt.proteinid.distinct())).select_from(Join)
    COUNT_PROTEIN = select(func.count(Protein.proteinid))
    # NPEP = select(func.count(PepProt.peptideid.distinct())).select_from(Join)
    NPEP = select(func.count(Peptide.peptideid))

    def __init__(
        self,
        peptide_filter: FormFilter | None = None,
        protein_filter: RowFilter
        | RowFilter
        | Sequence[tuple[str, str, Any]]
        | None = None,
        *,
        dtquery: DTQuery | None = None,
        columns: Sequence[Column | str] | None = None,
    ):
        self.dtquery = dtquery
        if isinstance(protein_filter, Sequence):
            protein_filter = RowFilter([RowQuery(*args) for args in protein_filter])

        self.protein_filter = protein_filter.to_sql(Protein) if protein_filter else None
        if peptide_filter is not None:
            self.peptide_filter, prot_where = peptide_filter.calc_where()
            if prot_where is not None:
                self.protein_filter = (
                    and_(prot_where, self.protein_filter)
                    if self.protein_filter is not None
                    else prot_where
                )

        else:
            self.peptide_filter = None
        if columns is None:  # pragma: no cover
            columns = []
        self.columns = [Column(c) if isinstance(c, str) else c for c in columns]

        self.search_columns = [c.name for c in self.columns if c.search]
        self.aggregates = [c.aggregate for c in self.columns if c.aggregate is not None]
        self.protein_columns = (
            [
                attr
                for col in self.columns
                if col.send and col.read
                for attr in [self.getcolumn(col)]
                if attr is not None
            ]
            if self.columns
            else [Protein]  # type: ignore
        )
        self._agg = [self.make_agg(agg) for agg in self.aggregates]

        npro = self.NPRO
        # qid = self.QID
        npep = self.NPEP
        if self.peptide_filter is not None:
            npro = npro.where(self.peptide_filter)  # type: ignore
            # qid = qid.where(self.peptide_filter)  # type: ignore
            npep = npep.where(self.peptide_filter)  # type: ignore
        self.npro = npro
        # self.qid = qid
        self.npep = npep

    def make_agg(self, agg: Aggregate) -> ColumnElement[Any]:
        return agg.make_column(Peptide)

    def protein_in_query(self) -> ColumnElement[Any]:
        assert self.peptide_filter is not None and len(self.aggregates) == 0
        return Protein.proteinid.in_(self.QID.where(self.peptide_filter))

    def peptide_subquery(self) -> Subquery:
        columns = [PepProt.proteinid, *self._agg]
        q = select(*columns).select_from(self.Join)
        if self.peptide_filter is not None:
            q = q.where(self.peptide_filter)
        q = q.group_by(PepProt.proteinid)
        return q.subquery()

    def getcolumn(self, col: Column) -> ColumnElement[Any]:
        attr = col.getcolumn(Protein)
        if attr is None or not hasattr(attr, "property"):
            raise ValueError(
                f"unknown column {col.name} for {Protein.__tablename__} table",
            )
        if isinstance(attr.property, Relationship):
            raise ValueError(f"column {col} is a relationship on protein!")
        return attr

    def build_queries(self) -> tuple[Select[Any], Select[Any]]:
        qid = None
        cols = []
        piq = None
        if self.aggregates:
            qid = self.peptide_subquery()
            cols = qid.columns[1:]
        elif self.peptide_filter is not None:
            piq = self.protein_in_query()

        columns = self.protein_columns + list(cols)

        q: Select[Any] = select(*columns)
        if qid is not None:
            q = q.join(qid, qid.c.proteinid == Protein.proteinid)
        elif piq is not None:
            q = q.where(piq)
        # if self.peptide_filter is not None:
        #     q = q.where(Protein.proteinid.in_(self.qid))
        if self.protein_filter is not None:
            q = q.where(self.protein_filter)

        q, tq = self.process_query(q)
        if qid is not None:
            tq = tq.join(qid, qid.c.proteinid == Protein.proteinid)

        return q, tq

    def query(
        self,
        engine: Engine,
        want_all: bool = True,
    ) -> ProteinQueryResult:
        q, tq = self.build_queries()
        total_proteins = total_peptides = total_filtered = -1
        with engine.begin() as conn:
            if want_all:
                total_proteins = conn.execute(self.npro).scalar() or 0
                total_peptides = conn.execute(self.npep).scalar() or 0
                total_filtered = conn.execute(tq).scalar() or 0
            df = pd.read_sql_query(q, con=conn)
            # sigh! sqllite does not have a std aggregate function
            # so we group_concat and the parse and the string again...
            for a in self.aggregates:
                if a.aggfunc in FAKE_AGGREGATE_FUNCS:
                    df[a.label] = df[a.label].apply(FAKE_AGGREGATE_FUNCS[a.aggfunc])

        return ProteinQueryResult(
            self.rehydrate(df),
            total_proteins,
            total_peptides,
            total_filtered,
        )

    def process_query(self, query: Select[Any]) -> tuple[Select[Any], Select[Any]]:
        q = self.dtquery
        if q:
            if q.search is not None and self.search_columns is not None:
                query = query.where(search_protein_tosql(q.search, self.search_columns))
            if q.order_column is not None:
                col = getattr(Protein, q.order_column, None)
                if col is not None:
                    query = query.order_by(col.asc() if q.ascending else col.desc())
                else:
                    # virtual
                    if q.order_column in {a.label for a in self.aggregates}:
                        asc = "ASC" if q.ascending else "DESC"
                        query = query.order_by(text(f"{q.order_column} {asc}"))

        filtered_total = self.COUNT_PROTEIN
        if query.whereclause is not None:
            filtered_total = filtered_total.where(query.whereclause)

        if q and q.length > 0:
            query = query.slice(q.start, q.start + q.length)
        return query, filtered_total

    def rehydrate(self, df: pd.DataFrame) -> pd.DataFrame:
        return df


## @export
class PeptideQuery:
    Join = join(PepProt, Protein)
    QID = select(PepProt.peptideid).select_from(Join)

    def __init__(
        self,
        peptide_filter: FormFilter | None = None,
        protein_filter: RowFilter | Sequence[tuple[str, str, Any]] | None = None,
        *,
        columns: Sequence[Column | str] | None = None,
        filtered_column: str | None = None,
    ):
        if isinstance(protein_filter, Sequence):
            protein_filter = RowFilter([RowQuery(*args) for args in protein_filter])
        self.protein_filter = protein_filter.to_sql(Protein) if protein_filter else None

        if peptide_filter is not None:
            self.peptide_filter, prot_where = peptide_filter.calc_where()
            if prot_where is not None:
                self.protein_filter = (
                    and_(prot_where, self.protein_filter)
                    if self.protein_filter is not None
                    else prot_where
                )

        else:
            self.peptide_filter = None

        self.filtered_column = filtered_column
        if columns is None:  # pragma: no cover
            columns = []
        self.columns = [Column(c) if isinstance(c, str) else c for c in columns]
        self.aggregates = [c.aggregate for c in self.columns if c.aggregate is not None]
        self._agg = [self.make_agg(c) for c in self.aggregates]
        self.peptide_columns = (
            [
                attr
                for col in self.columns
                if col.read
                for attr in [self.getcolumn(col)]
                if attr is not None
            ]
            if self.columns
            else [Peptide]  # type: ignore
        )

    def getcolumn(self, col: Column) -> ColumnElement[Any] | None:
        if col.aggregate is not None:
            return None
        ret = col.getcolumn(Peptide)
        if ret is None:
            raise ValueError(
                f"unknown column {col.name} for {Peptide.__tablename__} table",
            )
        return ret

    def peptide_in_query(self) -> ColumnElement[Any]:
        assert self.protein_filter is not None
        return Peptide.peptideid.in_(self.QID.where(self.protein_filter))

    def make_agg(self, agg: Aggregate) -> ColumnElement[Any]:
        return agg.make_column(Protein)

    def protein_subquery(self) -> Subquery:
        columns = [PepProt.peptideid, *self._agg]
        q = select(*columns).select_from(self.Join)
        if self.protein_filter is not None:
            q = q.where(self.protein_filter)
        q = q.group_by(PepProt.peptideid)
        return q.subquery()

    def count(self, engine: Engine) -> int:
        q = self.build_query()
        q = select(func.count()).select_from(q.subquery())  # type: ignore
        with engine.connect() as conn:
            return conn.execute(q).scalar_one()

    def build_query(
        self,
    ) -> Select[Any]:
        qid = None
        cols = []
        piq = None
        if self.aggregates:
            qid = self.protein_subquery()
            cols = qid.columns[1:]
        elif self.protein_filter is not None:
            piq = self.peptide_in_query()

        columns = self.peptide_columns + list(cols)

        if self.peptide_filter is not None and self.filtered_column is not None:
            fcol = self.peptide_filter.label(self.filtered_column)
            columns.append(fcol)  # type: ignore

        q: Select[Any] = select(*columns)
        if qid is not None:
            q = q.join(qid, qid.c.peptideid == Peptide.peptideid)
        elif piq is not None:
            q = q.where(piq)
        if self.peptide_filter is not None and self.filtered_column is None:
            q = q.where(self.peptide_filter)
        return q

    def query(
        self,
        engine: Engine,
    ) -> pd.DataFrame:
        q = self.build_query()

        with engine.begin() as conn:
            q = self.process_query(q, conn)
            df = pd.read_sql_query(q, con=conn)
        for a in self.aggregates:
            if a.aggfunc in FAKE_AGGREGATE_FUNCS:
                df[a.label] = df[a.label].apply(FAKE_AGGREGATE_FUNCS[a.aggfunc])
        return self.rehydrate(df)

    def process_query(self, q: Select[Any], conn: Connection) -> Select[Any]:
        return q

    def rehydrate(self, df: pd.DataFrame) -> pd.DataFrame:
        return rehydrate_peptides(df)


## @export
class SimplePeptideQuery:
    def __init__(
        self,
        peptide_filter: RowFilter | Sequence[tuple[str, str, Any]] | None = None,
        extra: FormFilter | None = None,
        *,
        columns: Sequence[Column | str] | None = None,
    ):
        if isinstance(peptide_filter, Sequence):
            peptide_filter = RowFilter([RowQuery(*args) for args in peptide_filter])

        self.peptide_filter = peptide_filter.to_sql(Peptide) if peptide_filter else None
        self.extra = extra
        if columns is None:  # pragma: no cover
            columns = []

        self.columns = [Column(col) if isinstance(col, str) else col for col in columns]

        self.peptide_columns = (
            [self.getcolumn(col) for col in self.columns if col.read]
            if self.columns
            else [Peptide]  # type: ignore
        )

    def getcolumn(self, col: Column) -> ColumnElement[Any]:
        if col.aggregate is not None:
            raise ValueError(f"no aggregate columns permitted: {col.name}")
        ret = col.getcolumn(Peptide)
        if ret is None:
            raise ValueError(
                f"unknown column {col.name} for {Peptide.__tablename__} table",
            )
        return ret

    def build_query(
        self,
    ) -> Select[Any]:
        q: Select[Any] = select(*self.peptide_columns)  # type: ignore

        if self.peptide_filter is not None:
            q = q.where(self.peptide_filter)
        if self.extra is not None:
            wc = self.extra.calc_where()
            if wc.pep_where is not None:
                q = q.where(wc.pep_where)
            # if wc.prot_where is not None:
            #     q = q.where(wc.prot_where)
        return q

    def query(
        self,
        engine: Engine,
    ) -> pd.DataFrame:
        q = self.build_query()
        with engine.connect() as conn:
            df = pd.read_sql_query(q, con=conn)
        return self.rehydrate(df)

    def rehydrate(self, df: pd.DataFrame) -> pd.DataFrame:
        return rehydrate_peptides(df)


## @export
@dataclass
class Stats:
    max_nnls_residual: float = 1.0
    max_maxPeakArea: float = 0.0
    min_enrichment: float = 0.0
    max_enrichment: float = 1.0


def up(v):
    return to_precision(v, 3, shift=1)


def down(v):
    return to_precision(v, 3, shift=-1)


ROUND: dict[str, Callable[[Any], Any]] = {
    "max_nnls_residual": up,
    "min_enrichment": down,
    "max_enrichment": up,
    "max_maxPeakArea": up,
}


## @export
def calc_stats(
    engine: Engine | None,
    round: bool = False,
    scale_intensity: bool = False,
) -> Stats:
    def default(v):  # pragma: no cover
        return to_precision(v, 3)

    if engine is None:  # pragma: no cover
        return Stats()
    if scale_intensity:  # pragma: no cover
        resid = Peptide.nnls_residual / Peptide.totalIntensityWeight
    else:
        resid = Peptide.nnls_residual / Peptide.totalNNLSWeight
    q = select(
        func.max(resid).label("max_nnls_residual"),
        func.max(Peptide.maxPeakArea).label("max_maxPeakArea"),
        func.min(Peptide.enrichment).label("min_enrichment"),
        func.max(Peptide.enrichment).label("max_enrichment"),
    )
    with engine.connect() as conn:
        row = conn.execute(q).mappings().one()
    if round:
        d = {k: ROUND.get(k, default)(v) for k, v in row.items()}
    else:
        d = dict(row)
    return Stats(**d)


## @export
def sqlite_version(engine: Engine | Path) -> int:
    from .model import Version

    dispose = False
    if isinstance(engine, Path):  # pragma: no cover
        engine = file2engine(engine)
        dispose = True

    q = select(
        func.max(Version.version).label(
            "version",
        ),
    )
    try:
        with engine.connect() as conn:
            row = conn.execute(q).mappings().one_or_none()
            if row is None:
                return -1
            return row["version"]
    except OperationalError:
        return 0
    finally:
        if dispose:
            engine.dispose()


def db_columns(engine: Engine | Path) -> tuple[list[str], list[str]]:
    from sqlalchemy import MetaData

    if isinstance(engine, Path):  # pragma: no cover
        engine = file2engine(engine)
    m = MetaData()
    try:
        m.reflect(only=["peptides", "proteins"], bind=engine)
        pep = m.tables["peptides"]
        prot = m.tables["proteins"]
        return [c.name for c in pep.columns], [c.name for c in prot.columns]

    except OperationalError:  # pragma: no cover
        return [], []



from __future__ import annotations

from base64 import b64encode
from io import BytesIO
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.figure import Figure
from scipy import stats

from .algo import settings_to_algo
from .logger import logger
from .sns import despine
from .utils import ensure_pos
from .utils import PeptideInfo
from .utils import resize

if TYPE_CHECKING:
    from matplotlib.axes import Axes
    from .utils import PeptideSettings
    from .types.turnovertype import TurnoverRow
    from .types.turnovertype import TurnoverDict

dpi = 96
FIGSIZE: tuple[float, float] = (1056.0 / dpi, 768.0 / dpi)

EICS_CMAP = plt.get_cmap("viridis")

NEICS_CMAP = len(EICS_CMAP.colors)  # type: ignore

EIC_COLUMNS = ["peptide", "modcol", "eics", "eics_shape", "monoFitParams"]


LABELLED_COLUMNS = [
    "labelledEnvelopes",
    "labelledElementCount",
    "peptide",
    "enrichment",
    "relativeIsotopeAbundance",
    "theoreticalDist",
]


FITTED_COLUMNS = [
    "peptide",
    "modcol",
    "isotopeEnvelopes",
    "maxIso",
    "relativeIsotopeAbundance",
    "enrichment",
    # "adjustedRsq",
]


ALL_COLUMNS = list(set(EIC_COLUMNS) | set(FITTED_COLUMNS) | set(LABELLED_COLUMNS))

# def resize_all(*arrs: np.ndarray) -> list[np.ndarray]:
#     mx = max([len(a) for a in arrs])
#     return [resize(a, mx) for a in arrs]


def pepsize(peptide: str, mx: int = 30) -> str:
    peptide = str(peptide)
    n = len(peptide)
    if n <= mx:
        return peptide
    n = m = (mx - 3) // 2
    n += mx - 2 * n - 3
    return f"{peptide[:n]}...{peptide[-m:]}"


def peptide_legend(peptide: TurnoverRow, enrichment: float | None = None) -> str:
    if enrichment is None:
        enrichment = peptide.enrichment
    if enrichment is not None:
        ie = round(peptide.enrichment * 100, 1)
        ia = round(peptide.relativeIsotopeAbundance * 100, 1)
        return f"{pepsize(peptide.peptide)}\nLPF: {ia}%\nEnrichment: {ie}%"
    else:  # pragma: no cover
        return f"{pepsize(peptide.peptide)},\nunable to estimate enrichment / LPF"


def despine_bottom(ax: Axes) -> None:
    despine(ax=ax, bottom=True, trim=True)


def despine_top(ax: Axes, offset: int = 5) -> None:
    despine(ax=ax, top=True, trim=True, offset=offset)


def plotEICS(
    peptide: TurnoverRow,
    ax: Axes | None = None,
    figsize: tuple[float, float] = FIGSIZE,
) -> Axes:
    from .fitenvelopes import normalIntensityFunction, FAIL_W_CURVE

    pep = peptide.peptide
    mod = peptide.modcol
    eics = peptide.eics
    monoFitParams = peptide.monoFitParams
    fail = len(monoFitParams) == 0  # old way
    if hasattr(peptide, "fail"):  # pragma: no cover
        fail = fail or bool(int(getattr(peptide, "fail")) & (1 << FAIL_W_CURVE))

    # maxInt = eics[0,:, 1].max()
    # ylim = (0, maxInt)
    if ax is None:  # pragma: no cover
        fig = Figure(figsize=figsize)
        ax = fig.subplots()
    ax.set(xlabel="retention time", ylabel="Intensity", title=f"{pepsize(pep)} {mod}")

    m = eics[0]
    rt, im = m[:, 0], m[:, 1]
    ax.plot(rt, im, color="dodgerblue", lw=2)
    oeics = eics[2:]
    step = max(NEICS_CMAP // len(oeics), 1) if len(oeics) else 1

    for idx, e in enumerate(oeics):
        rt, i = e[:, 0], e[:, 1]
        icolor = min(idx * step, NEICS_CMAP - 1)
        ax.plot(
            rt,
            i,
            color=EICS_CMAP.colors[icolor],  # type: ignore
            alpha=0.2,
            linestyle="dashed",
        )

    mono = eics[1]
    rt, im = mono[:, 0], mono[:, 1]
    ax.plot(rt, im, color="red", lw=2)
    d = dict(handlelength=0)
    if not fail:
        i = normalIntensityFunction(rt, *monoFitParams)
        ax.plot(rt, i, color="#8b2252", lw=2)  # violetred4
        ax.legend(labels=["Fitted"], labelcolor="darkgreen", **d)
    else:  # pragma: no cover
        ax.legend(labels=["Bad"], labelcolor="salmon", **d)

    return ax


def intrange(n: int) -> np.ndarray:
    return np.arange(0, n, 1, dtype=np.int32)


def binify(a: np.ndarray, bins: int) -> np.ndarray:
    r = stats.binned_statistic(
        np.arange(len(a)),
        a,
        statistic="sum",
        bins=bins,
    )
    return r.statistic


def plotLabelledEnvelope(
    peptide: TurnoverRow,
    settings: PeptideSettings,
    ax: Axes | None = None,
    scaled: bool = True,
    figsize: tuple[float, float] = FIGSIZE,
    *,
    enrichmentColumns: int = 0,
) -> Axes:
    if ax is None:  # pragma: no cover
        fig = Figure(figsize=figsize)
        ax = fig.subplots()
    if len(peptide.labelledEnvelopes) > 0:
        e = peptide.labelledEnvelopes
        e = e if isinstance(e, np.ndarray) else np.array(e)
        if enrichmentColumns and len(e) != enrichmentColumns:
            e = binify(e, enrichmentColumns)

        x = intrange(len(e))
        enrichments = settings.getEnrichmentsN(len(e))
        if scaled:
            total = e.sum()
            e = e / total
        ax.bar(x, e, color="turquoise", edgecolor="black")
        ax.set_xticks(x)
        ax.set_xticklabels(
            [f"{v:.2f}" for v in enrichments],
            rotation=-90,
        )
        ax.set(xlabel="enrichment")
        if scaled:
            ax.set(ylim=(0.0, 1.0), ylabel="fraction")

    else:  # pragma: no cover
        ie = peptide.labelledElementCount
        x = intrange(ie + 1)
        ax.bar(x, np.repeat(0, ie + 1), color="turquoise")
        ax.set_xticks(x)
    ax.legend(labels=[peptide_legend(peptide)], handlelength=0)
    despine_bottom(ax)
    ax.xaxis.set_ticks_position("none")
    ax.set(title="labelledEnvelopes")
    return ax


def plotFittedEnvelopes(
    peptide: TurnoverRow,
    settings: PeptideSettings,
    ax: Axes | None = None,
    figsize: tuple[float, float] = FIGSIZE,
    *,
    with_rsq: bool = False,
) -> Axes:
    if ax is None:  # pragma: no cover
        fig = Figure(figsize=figsize)
        ax = fig.subplots()

    isotopeEnvelopes = peptide.isotopeEnvelopes
    if len(isotopeEnvelopes) == 0:  # pragma: no cover
        logger.warning("plotFittedEnvelopes[%s]: no envelope!", peptide.peptide)
        return ax

    isotopeEnvelopes = resize(isotopeEnvelopes, peptide.maxIso + 2)

    theoreticalDist = np.array(peptide.theoreticalDist)
    natural_dist = settings_to_algo(settings).natural_dist
    naturalDist = natural_dist(
        PeptideInfo(peptide.peptide, settings),
        peptide.isotopeEnvelopes[1],  # monoScale
    )
    # naturalDist is of length settings.getElementCount(peptideFormula(peptide.peptide))

    naturalDist = resize(naturalDist, peptide.maxIso + 1)
    theoreticalDist = resize(theoreticalDist, peptide.maxIso + 1)

    x = intrange(len(isotopeEnvelopes)) - 1

    heavyDistribution = ensure_pos(isotopeEnvelopes[1:] - naturalDist)
    xm1, x1 = x[:1], x[1:]  # skip negative heavy atom
    ax.set(title=f"{pepsize(peptide.peptide)} {peptide.modcol}")
    ax.bar(
        x1,
        isotopeEnvelopes[1:],
        color="turquoise",
        alpha=1.0,
    )
    ax.bar(x1, heavyDistribution, color="blue", width=0.5, alpha=0.4)
    if with_rsq and hasattr(peptide, "adjustedRsq"):
        adjustedRsq = np.array(peptide.adjustedRsq)
        ax.bar(
            x1,
            adjustedRsq[1:] * isotopeEnvelopes[1],
            color="tomato",
            alpha=0.3,
            width=0.2,
        )

    ax.legend(labels=[peptide_legend(peptide)], handlelength=0, facecolor="none")
    ax.bar(xm1, isotopeEnvelopes[:1], color="red", alpha=0.3)
    ax.bar(x1, -theoreticalDist, color="#98F5FF", alpha=0.3)  # cadetblue1
    nd = resize(naturalDist, len(x) - 1)
    ax.vlines(x1, 0, nd, color="orange", linestyle="dashed")
    ax.scatter(x1, nd, color="orange", marker="o", s=20)

    xlabel = f"number of ${{}}^{{{settings.labelledIsotopeNumber}}}{settings.labelledElement}$ atoms"
    nfailed = np.isnan(isotopeEnvelopes[1:]).sum()
    if nfailed > 0:  # pragma: no cover
        xlabel += f" [{nfailed} failed fits]"
    ax.set(xlabel=xlabel)

    ax.set_xticks(x)

    despine_bottom(ax)
    ax.xaxis.set_ticks_position("none")

    return ax


def plotIntensities(
    peptide: TurnoverRow,  # eics
    eici: list[int] | None = None,
    ax: Axes | None = None,
    figsize: tuple[float, float] = FIGSIZE,
    *,
    with_origin: bool = False,
) -> Axes:
    from scipy.linalg import lstsq

    eics = peptide.eics
    mono_intensity = eics[1, :, 1]
    if with_origin:
        im = np.c_[np.ones(len(mono_intensity)), mono_intensity]
    else:
        im = mono_intensity.reshape(-1, 1)
    x = np.linspace(np.min(mono_intensity), np.max(mono_intensity), num=30)
    if eici is None:
        eici = list(range(2, eics.shape[0]))
    else:  # pragma: no cover
        idx = set(range(2, eics.shape[0]))
        eici = list(sorted(set(eici) & idx))
    if ax is None:  # pragma: no cover
        fig = Figure(figsize=figsize)
        ax = fig.subplots(1, 1)
    # step = max(NEICS_CMAP // len(eici), 1) if len(eici) else 1
    for i in eici:
        ints = eics[i, :, 1]
        _ret = lstsq(im, ints)
        assert _ret is not None
        fit, _residues, _rank, _s = _ret
        if with_origin:
            y = fit[0] + fit[1] * x
        else:
            y = fit[0] * x
        # icolor = min(i * step, NEICS_CMAP - 1)
        # color=EICS_CMAP.colors[icolor]
        ax.scatter(mono_intensity, ints, s=10)
        ax.plot(x, y)
    ax.set(
        title=f"{pepsize(peptide.peptide)} {peptide.modcol}",
        xlabel="mono intensity",
        ylabel="intensity",
    )
    return ax


## @export
def enrichment_plot(
    turn: pd.DataFrame,
    factor: float = 2.0,
    ax: Axes | None = None,
    figsize: tuple[float, float] | None = None,
    column: str = "enrichment",
) -> Axes:
    enrich = turn[column]
    enrich = enrich[enrich.notna()]
    em = enrich.mean()
    esd = enrich.std()
    if ax is None:
        fig = Figure(figsize=figsize if figsize else FIGSIZE)
        ax = fig.subplots(1, 1)

    x = np.linspace(enrich.min() - 3 * esd, enrich.max() + 3 * esd, 40)
    y = stats.norm(em, esd).pdf(x)  # type: ignore

    ax.hist(
        enrich,
        bins=40,
        color="turquoise",
        alpha=0.4,
        density=True,
        edgecolor="#999",
    )
    ax.plot(x, y)
    ax.vlines([em - factor * esd, em + factor * esd], 0, [2.5, 2.5], color="red")
    despine_top(ax, offset=5)
    return ax


def nnls_values(  # pragma: no cover
    df: pd.DataFrame,
    percent: float = 0.01,
) -> pd.Series:
    nnls_residual = df["nnls_residual"]
    if "totalNNLSWeight" in df.columns:
        nnls_residual = nnls_residual / df["totalNNLSWeight"]
    elif "totalIntensityWeight" in df.columns:
        nnls_residual = nnls_residual / df["totalIntensityWeight"]
    nnls_residual = nnls_residual[nnls_residual.notna()]
    if percent == 0.0:
        return nnls_residual

    bins, bounds = np.histogram(nnls_residual, bins=500)

    maxval = bins.max()
    imax = np.argmax(bins == maxval)
    if maxval > 0:
        imin = np.argmin(bins[imax:] / maxval > percent)
    else:
        imin = 0

    idx = min(len(bounds) - 1, int(imax + imin))

    return nnls_residual[nnls_residual < bounds[idx]]


## @export
def nnls_plot(  # ['nnls_residual', 'totalNNLSWeight', 'totalIntensityWeight']
    df: pd.DataFrame,
    ax: Axes | None = None,
    figsize: tuple[float, float] | None = None,
    percent: float = 0.01,
) -> Axes:
    # nnls = nnls_values(df, percent=percent)
    nnls = df["nnls_residual"]

    if ax is None:
        fig = Figure(figsize=figsize if figsize else FIGSIZE)
        ax = fig.subplots(1, 1)
    # mn,mx = nnls_residual.min(), nnls_residual.max()
    ax.hist(nnls, bins="auto", color="turquoise")
    despine(ax, trim=True, offset=5)
    ax.set(ylabel="counts", xlabel="NNLS deviance")
    return ax


def correlation_plot(
    df: pd.DataFrame,
    column: str,
    xlabel: str | None = None,
    ax: Axes | None = None,
    figsize: tuple[float, float] | None = None,
) -> Axes:
    if ax is None:
        fig = Figure(figsize=figsize if figsize else FIGSIZE)
        ax = fig.subplots(1, 1)
    # mn,mx = nnls_residual.min(), nnls_residual.max()
    col = df[column]
    ax.hist(col, bins="auto", color="turquoise")
    despine(ax, trim=True, offset=5)
    ax.set(ylabel="counts", xlabel=xlabel or column)
    return ax


## @export
def plot_all(
    peptide: TurnoverRow,
    settings: PeptideSettings,
    figsize: tuple[float, float] | None = None,
    hspace: float = 0.5,
    *,
    layout: tuple[int, int] = (1, 3),
    enrichmentColumns: int = 0,
) -> tuple[Figure, list[Axes]]:
    r, c = layout
    if figsize is None:
        # width, height
        figsize = c * FIGSIZE[0], r * FIGSIZE[1]

    fig = Figure(figsize=figsize)
    axes = fig.subplots(r, c, gridspec_kw=dict(hspace=hspace), squeeze=False).flatten()

    plotEICS(peptide, ax=axes[0])
    plotFittedEnvelopes(peptide, settings, ax=axes[1])
    plotLabelledEnvelope(
        peptide,
        settings,
        ax=axes[2],
        enrichmentColumns=enrichmentColumns,
    )
    if len(axes) > 3:
        from .config import WITH_ORIGIN

        plotIntensities(peptide, None, ax=axes[3], with_origin=WITH_ORIGIN)
    return fig, axes  # type: ignore


def rename(s: pd.Series, suffix: str = "_me") -> pd.Series:  # pragma: no cover
    r = pd.Series(dtype=object)
    for key in [
        "isotopeEnvelopes",
        "relativeIsotopeAbundance",
        "monoFitParams",
        "eics",
        "labelledEnvelopes",
        "labelledElementCount",
        "enrichment",
        "heavyFitParams",
        "heavyCor",
    ]:
        r[key] = s[key + suffix]

    r.peptide = s.peptide
    r.modcol = s.modcol
    r.maxIso = int(s["maxIso" + suffix])
    return r


def plot_eics_sidebyside(  # pragma: no cover
    b: TurnoverRow,
    suffix: str = "_me",
    figsize: tuple[float, float] = (15.0, 5.0),
) -> tuple[Figure, Axes, Axes]:
    fig = Figure(figsize=figsize)
    [ax1, ax2] = fig.subplots(1, 2)

    plotEICS(b, ax=ax1)
    plotEICS(rename(b, suffix), ax=ax2)  # type: ignore
    return fig, ax1, ax2


def plot_fitted_sidebyside(  # pragma: no cover
    b: TurnoverRow,
    settings: PeptideSettings,
    suffix: str = "_me",
    figsize: tuple[float, float] = (15, 5),
) -> tuple[Figure, Axes, Axes]:
    fig = Figure(figsize=figsize)
    [ax1, ax2] = fig.subplots(1, 2)

    plotFittedEnvelopes(b, settings, ax=ax1)
    plotFittedEnvelopes(rename(b, suffix), settings, ax=ax2)  # type: ignore
    return fig, ax1, ax2


def plot_labelled_sidebyside(  # pragma: no cover
    b: TurnoverRow,
    settings: PeptideSettings,
    suffix: str = "_me",
    figsize: tuple[float, float] = (15, 5),
) -> tuple[Figure, Axes, Axes]:
    fig = Figure(figsize=figsize)
    [ax1, ax2] = fig.subplots(1, 2)

    plotLabelledEnvelope(b, settings, ax=ax1)
    plotLabelledEnvelope(rename(b, suffix), settings, ax=ax2)  # type: ignore
    return fig, ax1, ax2


def plot_fit_arrays(  # pragma: no cover
    row: TurnoverRow,
    suffix: str = "_me",
    figsize: tuple[float, float] = (15, 12),
) -> tuple[Figure, list[list[Axes]]]:
    b: TurnoverDict = row  # type: ignore
    fig = Figure(figsize=figsize)
    axes = fig.subplots(
        5,
        2,
        gridspec_kw=dict(hspace=0.5),
    )
    [[ax1, ax2], [ax3, _ax4], [ax5, ax6], [_ax7, ax8], [ax9, _ax10]] = axes

    def plot_a(name: str, ax: Axes) -> None:
        x = list(range(0, len(b[name])))  # type: ignore
        v = b[name + suffix]  # type: ignore
        if len(x) != len(v):
            ax.set(title=f"bad {name}")
            return
        dodge = np.nanmax(np.abs(v)) / 50
        ax.plot(x, b[name])  # type: ignore
        ax.plot(x, v + dodge)
        ax.set(title=f"{name}")

    def plot_a_col(name: str, ax: Axes, col: int) -> None:
        x = list(range(0, len(b[name][:, col])))  # type: ignore
        v = b[name + suffix][:, col]  # type: ignore
        if len(x) != len(v):
            ax.set(title=f"bad {name}")
            return
        dodge = np.nanmax(np.abs(v)) / 50
        ax.plot(x, b[name][:, col])  # type: ignore
        ax.plot(x, v + dodge)
        ax.set(title=f"{name}")

    def ratio(name: str) -> np.ndarray:
        a = b[name]  # type: ignore
        m = b[name + suffix]  # type: ignore
        if len(a) != len(m):
            return a * 0
        both = (a == 0) & (m == 0)
        ret = m / np.where(a != 0, a, 1e-6)
        if both.sum() > 0:  # 0/0 is 1 !
            ret = np.where(both, 1.0, ret)
        return ret

    def plot_ratio(name: str, ax: Axes) -> None:
        val = b[name]  # type: ignore
        x = list(range(0, len(val)))
        ax.plot(x, ratio(name))
        ax.set(title=f"{name} ratio")

    plot_a_col("isotopeRegressions", ax1, 1)
    ax1.set(title="regression intensity")

    plot_a_col("isotopeRegressions", ax2, 0)
    ax2.set(title="adjusted R-squared")

    plot_a("isotopeEnvelopes", ax3)
    # plot_a("heavyDistribution", ax4)
    # plot_heavy_fit(ax4)
    plot_a("labelledEnvelopes", ax5)
    plot_a("monoFitParams", ax6)
    # plot_a("heavyFitParams", ax7)

    plot_ratio("monoFitParams", ax8)
    plot_ratio("isotopeEnvelopes", ax9)
    # plot_ratio("heavyDistribution", ax10)

    return fig, axes


# for use in jupyter notebooks....
def plot_all_sidebyside(  # pragma: no cover
    b: TurnoverRow,
    settings: PeptideSettings,
) -> tuple[Figure, Figure, Figure, Figure]:
    f1, *_axes = plot_eics_sidebyside(b)
    f2, *_axes = plot_fitted_sidebyside(b, settings)
    f3, *_axes = plot_labelled_sidebyside(b, settings)
    f4, _axes = plot_fit_arrays(b)
    return f1, f2, f3, f4


## @export
def to_image_url(fig: Figure, format: str = "png") -> str:
    out = BytesIO()
    fig.savefig(out, format=format)
    data = f"data:image/{format.lower()};base64,".encode("ascii") + b64encode(
        out.getbuffer(),
    )
    return data.decode("ascii")


## @export
def plotLPF(
    df: pd.DataFrame,  # ['peptide', 'relativeIsotopeAbundance']
    ax: Axes | None = None,
    figsize: tuple[float, float] = (8.0, 1.5),
    sort: bool = True,
) -> Axes:
    if ax is None:
        fig = Figure(figsize=figsize)
        ax = fig.subplots()
    d = df.set_index("peptide")
    d = d["relativeIsotopeAbundance"]
    if sort:
        d = d.sort_values()
    # d.plot(ax=ax, style=".-", color="turquoise")
    d.plot(ax=ax, kind="bar", color="turquoise", edgecolor="black")
    ax.set(ylabel="LPF", ylim=(0, 1), xticks=[])
    despine(
        ax=ax,
        top=True,
        bottom=False,
        trim=True,
        offset=0,
    )
    legend = ax.get_legend()
    if legend is not None:  # pragma: no cover
        legend.remove()

    fig = ax.get_figure()
    if hasattr(fig, "tight_layout"):
        fig.tight_layout()  # type: ignore
    return ax


def plotLogSigmaLPF(
    df: pd.DataFrame,
    ax: Axes | None = None,
    figsize: tuple[float, float] = (8.0, 4.0),
    remove_zero_std: bool = True,
) -> Axes:
    if ax is None:
        fig = Figure(figsize=figsize)
        ax = fig.subplots()
    key = df["lpf_median"] != 0.0
    if remove_zero_std:
        key = key & (df["lpf_std"] != 0.0)
    d = df[key]
    x = np.log((d["lpf_std"] + 1e-4) / (d["lpf_median"]))
    assert isinstance(x, pd.Series)
    x.plot.hist(ax=ax, bins=100, color="turquoise", grid=False)
    despine(ax, trim=True, offset=5)
    ax.set(ylabel="counts", xlabel=r"$log(\sigma/LPF)$")
    return ax



from __future__ import annotations

from . import background_ui
from . import pymz_ui
from . import web_ui
from .cli import cli
from .sqla import iso_peaks_ui

__all__ = ["background_ui", "pymz_ui", "iso_peaks_ui", "web_ui", "cli"]

# pylint: disable=unused-import


# REM: imports should *not* import scipy,pandas,numpy!!!! Just click and .cli
# and a few system modules os, sys, typing... BE CAREFUL! otherwise load
# times might slow user interaction.

if __name__ == "__main__":  # pragma: no cover
    cli.main(prog_name="turnover")



from __future__ import annotations

import logging
import mmap
import re
from dataclasses import asdict
from datetime import datetime
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Iterator
from typing import NamedTuple
from typing import overload
from typing import TYPE_CHECKING
from typing import TypedDict

import numpy as np
import pandas as pd
from typing_extensions import override

from .fitenvelopes import FAIL_E_NOEICS
from .fitenvelopes import fitEnvelope
from .jobs import TurnoverJob
from .logger import logger
from .logger import show_log
from .parallel_utils import Task
from .resourcefiles import BaseResourceFile
from .resourcefiles import MzMLResourceFile
from .resourcefiles import MzMLResourceFileLocal
from .resourcefiles import PepXMLResourceFile
from .resourcefiles import ProtXMLResourceFile
from .resourcefiles import ResourceFiles
from .resourcefiles import ResultsResourceFile
from .types.pepxmltypes import PepXMLRunRowRT
from .utils import array_split
from .utils import getsize
from .utils import human
from .utils import IO
from .utils import PeptideSettings

if TYPE_CHECKING:
    from pymzml.spec import Spectrum


class RTDict(TypedDict):
    retention_time_sec: np.ndarray
    scanindex: np.ndarray
    mzmax: np.ndarray
    mzmin: np.ndarray
    imax: np.ndarray


SPECTRUM_QUERY = re.compile(b"<spectrum ")
CENTRIODED = re.compile(b'accession="MS:1000127"')


## @export
def scan_mzml_spectra(mzml: Path) -> Iterator[int]:
    """Very fast inspection of .mzML spectra queries"""
    with mzml.open("rb") as fp:
        with mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            for m in SPECTRUM_QUERY.finditer(mm):
                yield m.start(0)


def maybe_centroided_spectra(mzml: Path) -> bool:
    """Very fast inspection of .mzML spectra queries"""
    with mzml.open("rb") as fp:
        with mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            for m in CENTRIODED.finditer(mm):
                return True
    return False


class Search(NamedTuple):
    mzmin: float
    mzmax: float
    rtmin: float
    rtmax: float

    def isoverlap(self, other: Search) -> bool:
        dmz = min(self.mzmax, other.mzmax) - max(self.mzmin, other.mzmin)
        if dmz < 0:
            return False
        drt = min(self.rtmax, other.rtmax) - max(self.rtmin, other.rtmin)
        if drt < 0:
            return False
        return True


class ByteOrderError(ValueError):
    pass


def mzi_max_intensity(
    mzi: np.ndarray,
    mzmin: float,
    mzmax: float,
    centroided: bool = True,
) -> float:
    mz, i = mzi[0], mzi[1]
    if mzmin > mz[-1] or mzmax < mz[0]:  # pragma: no cover
        # completly out of range
        return 0.0
    q = (mz <= mzmax) & (mz >= mzmin)
    if q.any():
        return i[q].max()
    # we are wholly *between* two mz points so we linearly interpolate and then find max.
    # *NO* centroided peaks mean voids between peaks are basically zero....
    if centroided:
        return 0.0

    iv = np.interp(np.array([mzmin, mzmax]), mz, i)
    return iv.max()


def zeroi() -> np.ndarray:
    return np.array([], dtype=np.float32).reshape(MZI.SHAPE)


EIC_COLUMNS = ("mzmin", "mzmax", "rtmin", "rtmax")
EIC_COLUMNS2 = ("mzranges", "rtmin", "rtmax")


class MZI:
    SHAPE = (2, -1)

    def __init__(
        self,
        mzmlfile: MzMLResourceFile,
    ):
        self.name = mzmlfile.name
        meta, mapname = mzmlfile.cache_mzml(), mzmlfile.cache_memmap()

        self._mzi = np.memmap(mapname, dtype=np.float32, mode="r")

        if self._mzi[0] != MAGIC_NO:
            raise ByteOrderError(
                f'file "{mapname}" can\'t be read (written with different byteorder)',
            )
        # columns: retention_time_sec', 'scanindex', 'mzmax', 'mzmin', 'imax'
        self.df = IO(meta).read_df()

        scanstart = np.concatenate(  # pylint: disable=unexpected-keyword-arg
            ([1], self.df["scanindex"].to_numpy(dtype=np.int64)[:-1]),
            dtype=np.int64,
        )
        self.df["scanstart"] = scanstart
        mzmax = self.df["mzmax"].max()
        mzmin = self.df["mzmin"].min()
        rtmin = self.df["retention_time_sec"].min()
        rtmax = self.df["retention_time_sec"].max()
        self.search = Search(mzmin, mzmax, rtmin, rtmax)
        self.centroided = True

        self.dino = None
        # dino = mzmlfile.cache_dinosaur()
        # if dino.exists():
        #     self.dino = IO(dino).read_df(columns=["mz", "rtStart", "rtEnd"])
        # else:
        #     self.dino = None

    def __repr__(self) -> str:
        return (
            f"MZI({self.name}[{len(self.df)}],"
            f" mz=[{self.search.mzmin:.2f},{self.search.mzmax:.2f}],"
            f" rt=[{self.search.rtmin:.2f},{self.search.rtmax:.2f}])"
        )

    def mzi_column(self, df: pd.DataFrame | None = None) -> pd.Series:
        if df is None:  # pragma: no cover
            df = self.df
        return df[["scanstart", "scanindex"]].apply(self.mzi_apply, axis=1)

    def mzi(self, scanstart: int, scanindex: int) -> np.ndarray:
        return self._mzi[scanstart:scanindex].reshape(self.SHAPE)

    def mzi_apply(self, row: pd.Series) -> np.ndarray:
        return self.mzi(row["scanstart"], row["scanindex"])

    @classmethod
    def from_mzml_file(
        cls,
        mzmlfile: str | Path,
        cache_dir: str | Path | None = None,
        *,
        quiet: bool = False,
    ) -> MZI:
        res = MzMLResourceFileLocal(mzmlfile, cache_dir=cache_dir)
        if not res.cache_mzml_ok():  # pragma: no cover
            if quiet:
                mzml_create(res)
            else:
                with show_log():
                    mzml_create(res)
        return cls(res)

    @classmethod
    def from_mzi_file(
        cls,
        path: str | Path,
        cache_dir: str | Path | None = None,
    ) -> MZI:
        path = Path(path)
        if not path.exists():
            raise FileNotFoundError(f'"{path}" doesn\'t exist!')
        res = MzMLResourceFileLocal(path, cache_dir=cache_dir)
        return cls(res)

    @classmethod
    def can_read_mzi(cls, mzi: str | Path) -> bool:
        if not Path(mzi).exists():
            return False
        try:
            _mzi = np.memmap(mzi, dtype=np.float32, mode="r")
            return _mzi[0] == MAGIC_NO
        except ValueError:  # pragma: no cover
            return False

    def getEIC(
        self,
        mzrange: np.ndarray,  # float[maxIso+2,2]
        rtmin: float,
        rtmax: float,
        *,
        minrt: int = 0,
    ) -> np.ndarray | None:  # float[maxIso+2,len(rt),2]
        # ensure within boundary
        s = self.search
        if not (
            rtmin >= s.rtmin
            and rtmax <= s.rtmax
            and np.any(mzrange[:, 0] <= s.mzmax)
            and np.any(mzrange[:, 1] >= s.mzmin)
        ):
            return None
        rt = self.df["retention_time_sec"]
        q = (rt >= rtmin) & (rt <= rtmax)
        df = self.df[q]
        if len(df) < minrt:  # pragma: no cover
            return None

        mzi = self.mzi_column(df)
        rt = df["retention_time_sec"]
        out = np.zeros((len(mzrange), len(rt), 2), dtype=np.float32)
        for idx, (mzmin, mzmax) in enumerate(mzrange):
            # xcms foobar! XCMS_STEP = 0.1
            # if XCMS_STEP > 0.0:
            #     mzmin, mzmax = (
            #         floor(mzmin / XCMS_STEP) * XCMS_STEP - XCMS_STEP / 2,
            #         ceil(mzmax / XCMS_STEP) * XCMS_STEP + XCMS_STEP / 2,
            #     )
            outofrange = (df["mzmax"] < mzmin) | (df["mzmin"] > mzmax)
            iz = mzi[outofrange].apply(lambda i: 0.0)
            inz = mzi[~outofrange].apply(
                mzi_max_intensity,
                args=(mzmin, mzmax, self.centroided),
            )
            # stack columns and reset index to rt
            imax = pd.concat(
                [idf for idf in [iz, inz] if len(idf) > 0],
                axis=0,
            ).reindex(index=rt.index)
            # 2 column array [rt, intensity]
            rti = pd.concat([rt, imax], axis=1)
            out[idx, :, :] = rti.to_numpy(dtype=np.float32)

        if out[:, :, 1].max() <= 0.0:  # pragma: no cover
            return None
        return out

    def subset(
        self,
        mzmin: float,
        mzmax: float,
        rtmin: float,
        rtmax: float,
    ) -> pd.DataFrame:
        """data from range mzmin <= mz <= mzmax and rtmin <= retention_time_sec <= rtmax"""
        rt = self.df["retention_time_sec"]
        q = (
            (rt >= rtmin)
            & (rt <= rtmax)
            & (self.df["mzmin"] <= mzmax)
            & (self.df["mzmax"] >= mzmin)
        )
        if q.any():
            return self.df[q]
        return pd.DataFrame({n: [] for n in self.df.columns})

    def _getEIC(
        self,
        mzmin: float,
        mzmax: float,
        rtmin: float,
        rtmax: float,
    ) -> np.ndarray | None:
        if not self.search.isoverlap(Search(mzmin, mzmax, rtmin, rtmax)):
            return None

        rt = self.df["retention_time_sec"]
        q = (rt >= rtmin) & (rt <= rtmax)
        # t = (
        #     (rt >= rtmin)
        #     & (rt <= rtmax)
        #     & (self.df["mzmin"] <= mzmax)
        #     & (self.df["mzmax"] >= mzmin)
        # )
        if q.sum() == 0:
            return None

        filtered_df = self.df[q]

        def toimax(row: pd.Series) -> float:
            return mzi_max_intensity(self.mzi_apply(row), mzmin, mzmax, self.centroided)

        imax = filtered_df[["scanstart", "scanindex"]].apply(
            toimax,
            axis=1,
        )

        if imax.max() <= 0.0:  # pragma: no cover
            return None

        rti = pd.concat([filtered_df["retention_time_sec"], imax], axis=1)
        # array will be in column major 'F' order so
        # that transpose will be ok
        return rti.to_numpy(dtype=np.float32)

    def imax(self, df: pd.DataFrame, mzmin: float, mzmax: float) -> np.ndarray:
        df = df[["scanstart", "scanindex"]]
        imax = []
        for r in df.itertuples():
            mz, i = self.mzi(r.scanstart, r.scanindex)  # type: ignore
            q = (mzmin <= mz) & (mzmax >= mz)
            if q.any():
                m = i[q].max()
                imax.append(m)
            else:
                if self.centroided:
                    m = 0.0
                else:
                    m = np.interp(np.array([mzmin, mzmax]), mz, i).max()
                imax.append(m)

        return np.array(imax, dtype=np.float32)

    def eics_from_df(
        self,
        df: pd.DataFrame,
        columns: list[str] | None = None,
        *,
        zero_length: bool = False,
    ) -> pd.Series:
        if columns:
            df = df.rename(
                columns=dict(zip(columns, EIC_COLUMNS)),
            )

        mzmin, mzmax, rtmin, rtmax = EIC_COLUMNS
        _getEIC = self._getEIC

        def findeic(row: pd.Series) -> np.ndarray | float:
            ret = _getEIC(row[mzmin], row[mzmax], row[rtmin], row[rtmax])
            if ret is None:
                return zeroi() if zero_length else np.nan
            return ret.T

        return df[list(EIC_COLUMNS)].apply(findeic, axis=1)

    def eics_from_mzrange_df(
        self,
        df: pd.DataFrame,
        columns: list[str] | None = None,
        *,
        zero_length: bool = False,
    ) -> pd.Series:
        if columns:
            df = df.rename(
                columns=dict(zip(columns, EIC_COLUMNS2)),
            )
        mzranges, rtmin, rtmax = EIC_COLUMNS2
        _getEIC = self.getEIC

        def findeic(row: pd.Series) -> np.ndarray | float:
            ret = _getEIC(row[mzranges], row[rtmin], row[rtmax])
            if ret is None:
                return zeroi().reshape((-1, 0, 2)) if zero_length else np.nan
            return ret

        return df[list(EIC_COLUMNS2)].apply(findeic, axis=1)

    @classmethod
    def dehydrate_eics(cls, s: pd.Series) -> pd.Series:
        def f(a: Any) -> Any:
            return a.flatten() if isinstance(a, np.ndarray) else a

        return s.apply(f)

    @classmethod
    def rehydrate_eics(cls, s: pd.Series) -> pd.Series:
        def f(a: Any) -> Any:
            return a.reshape(cls.SHAPE) if isinstance(a, np.ndarray) else a

        return s.apply(f)

    def __len__(self) -> int:
        return len(self.df)

    @overload
    def __getitem__(self, idx: int) -> np.ndarray: ...

    @overload
    def __getitem__(self, idx: slice) -> pd.Series: ...

    @overload
    def __getitem__(self, idx: list[int]) -> pd.Series: ...

    def __getitem__(self, idx: int | list[int] | slice) -> np.ndarray | pd.Series:
        df = self.df[["scanstart", "scanindex"]].iloc[idx]
        if isinstance(idx, int):
            scanstart: int
            scanindex: int
            scanstart, scanindex = df  # type: ignore
            return self.mzi(scanstart, scanindex)

        return df.apply(self.mzi_apply, axis=1)  # type: ignore


# number must survive roundtrip....
# struct.unpack("f", struct.pack("f", MAGIC_NO))[0] == MAGIC_NO
MAGIC_NO = -0.1234000027179718
# MAGIC_BYTES = b'$\xb9\xfc\xbd'
MAGIC_BYTES = np.array([MAGIC_NO], dtype=np.float32).tobytes()


def centroided(spectrum: Spectrum) -> tuple[np.ndarray, np.ndarray]:
    p = spectrum.peaks("centroided")
    assert p is not None and isinstance(p, np.ndarray) and p.shape[-1] == 2, p
    mz, i = p.T
    return mz, i


def mzml_create(
    mzml: MzMLResourceFile,
    level: int = 0,
    number_of_bg_processes: int = 1,
    *,
    ensure_centroided: bool = False,
) -> int:
    """Create cache files for a mzML file"""
    from pymzml.spec import Spectrum, Chromatogram
    from .broken_api import PyMzMLReader
    from .logger import log_iterator

    mzreader = PyMzMLReader(
        mzml.original,
        build_index_from_scratch=False,
    )
    total = mzreader.get_spectrum_count()

    rt_dict = RTDict(
        retention_time_sec=np.zeros(total, dtype=np.float32),
        scanindex=np.zeros(total, dtype=np.int64),
        mzmin=np.zeros(total, dtype=np.float32),
        mzmax=np.zeros(total, dtype=np.float32),
        imax=np.zeros(total, dtype=np.float32),
    )
    n = 0
    scanindex = 1  # first part is a magic number
    mname = mzml.cache_memmap()

    mzml.ensure_cache_dir()

    with mname.open("wb") as fp:
        it: Iterator[Spectrum | Chromatogram] = log_iterator(
            mzreader,
            total=total,
            desc=mzml.original.name,
            level=level,
            number_of_bg_processes=number_of_bg_processes,
        )

        fp.write(MAGIC_BYTES)
        mz: np.ndarray
        i: np.ndarray
        # loop over <spectrumList count="nnnn"><spectrum>....</spectrum>... </spectrumList>

        for spectrum in it:
            # <cvParam cvRef="MS" accession="MS:1000511" name="ms level" value="1"/>
            assert isinstance(spectrum, Spectrum)
            if spectrum.ms_level == 1:
                # <cvParam cvRef="MS" accession="MS:1000016" name="scan start time" value="0.001483586933"
                #       unitCvRef="UO" unitAccession="UO:0000031" unitName="minute"/>

                stm = spectrum.scan_time_in_minutes()
                assert isinstance(stm, float)
                rt = stm * 60.0
                if ensure_centroided and not spectrum.get("MS:1000127"):
                    # raise ValueError("expecting centroided spectra!")
                    mz, i = centroided(spectrum)
                else:
                    mz, i = spectrum.mz, spectrum.i

                mzi = np.vstack((mz, i))
                # mz first, intensity second
                data = mzi.flatten().astype(np.float32)  # .tobytes()
                scanindex += len(data)
                # we will use these to search for suitable rt,mz from pep.xml files
                rt_dict["retention_time_sec"][n] = rt
                rt_dict["mzmax"][n] = mz.max()
                rt_dict["mzmin"][n] = mz.min()

                rt_dict["imax"][n] = i.max()
                rt_dict["scanindex"][n] = scanindex
                n += 1
                fp.write(data.tobytes())

    df = pd.DataFrame({k: v[:n] for k, v in rt_dict.items()})  # type: ignore
    df.sort_values(by="retention_time_sec", inplace=True)
    df.reset_index(drop=True, inplace=True)

    mzml_out(mzml, df)

    from .dinosaur.dinosaur import mzml_dinosaur

    mzml_dinosaur(mzml)

    return level


def getmzml(mzml: MzMLResourceFile) -> MZI:
    if not mzml.cache_ok():
        mzml_create(mzml)
        assert mzml.cache_ok()
    return MZI(mzml)


def mzml_out(
    mzml: MzMLResourceFile,
    df: pd.DataFrame,
) -> None:
    out = mzml.cache_mzml()
    IO(out, df).save_df()

    if logger.isEnabledFor(logging.INFO):  # pragma: no cover
        mname = mzml.cache_memmap()
        mem = df.memory_usage(deep=True).sum()
        g = getsize
        msg = (
            f"written: {mzml.original.name} -> {out.name}: memory={human(mem)} ondisk={human(g(out))},"
            f" mzimap={human(g(mname))} original={human(g(mzml.original))}"
        )
        logger.info(msg)


def dfinfo(df: pd.DataFrame, mzml: MzMLResourceFile, out: Path) -> None:
    if logger.isEnabledFor(logging.INFO):
        mem = df.memory_usage(deep=True).sum()
        g = getsize
        ret = f"{out.name}: memory={human(mem)} ondisk={human(g(out))} original={human(g(mzml.original))}"
        logger.info(ret)


def dehydrate_eics(df: pd.DataFrame) -> pd.DataFrame:
    df["eics_shape"] = df["eics"].apply(lambda a: a.shape)
    df["eics"] = df["eics"].apply(lambda e: e.flatten())
    return df


def turnover_run(
    job: TurnoverJob,
    jobrun: ResultsResourceFile,
    *,
    workers: int = 4,
    save_subset: bool = True,
    nspectra: int | None = None,
    cleanup: bool = True,
    save_job: bool = False,
    testing: int | None = None,
    compress_result: bool = False,
) -> None:
    from .pepxml import get_and_filter_pepxml, dehydrate_pepxml
    from .utils import human, rmfiles, compress
    from .parallel_utils import parallel_tasks

    start = datetime.now()
    pepxml_df = get_and_filter_pepxml(job)

    tasks = create_mz_tasks(pepxml_df, job, nspectra=nspectra, testing=testing)
    if logger.isEnabledFor(logging.INFO):
        taskmem = sum(t.mem() for t in tasks)
        mem = pepxml_df.memory_usage(deep=True).sum()
        info = f"pepxml memory={human(mem)}, tasks[{len(tasks)}/{workers}] total memory=[{human(taskmem)}]"
        logger.info(info)
    # save
    if save_subset:
        pepxml_df = pd.concat(
            [pepxml_df.loc[task.pepxml_df.index] for task in tasks],
            axis="index",
        )
        pepxml_df.index.name = "pepxml_index"
        pepxml_df = pepxml_df.reset_index()
    pepxml_df_filename = jobrun.cache_pepxml()
    pepxml_df = dehydrate_pepxml(pepxml_df)  # dehydrate the mzranges
    IO(pepxml_df_filename, pepxml_df).save_df()
    pepxml_df = None
    # get mzml tasks
    envelopes: list[Path] = []

    ntotal = len(tasks)
    try:
        for idx, (df, mzml) in enumerate(parallel_tasks(tasks, workers=workers), 1):
            logger.info(
                "EIC[%s] task done[%d/%d]: %s[%d]",
                jobrun.original.name,
                idx,
                ntotal,
                mzml.original.name,
                len(df),
            )
            if len(df) == 0:  # pragma: no cover
                continue

            eicsfile = jobrun.cache_partial_envelope(idx)
            df = dehydrate_eics(df)
            IO(eicsfile, df).save_df()
            envelopes.append(eicsfile)

            dfinfo(df, mzml, eicsfile)

        sr = SaveResult(job, jobrun, envelopes)
        sr.save()
        if compress_result:
            logger.info("compressing result file %s", jobrun.result_file().name)
            compress(jobrun.result_file(), remove=True)
        if save_job:
            toml = jobrun.result_file().with_suffix(".toml")
            job.save_file(toml)
        logger.info("turnover finished! %s after %s", job.jobid, datetime.now() - start)
    finally:
        if cleanup:
            envelopes.append(pepxml_df_filename)
            rmfiles(envelopes)


class SaveResult:
    def __init__(
        self,
        job: TurnoverJob,
        jobrun: ResultsResourceFile,
        eics: list[Path],
    ) -> None:
        self.job = job
        self.jobrun = jobrun
        self.eics = eics

    def save(self) -> None:
        rf = self.job.to_resource_files()
        self.save_result(
            self.jobrun.result_file(),
            self.jobrun.cache_pepxml(),
            rf.protxml,
            self.eics,
        )

    def save_result(
        self,
        result_name: Path,
        pepdf: Path,
        protdf: ProtXMLResourceFile,
        envelopes: list[Path],
    ) -> None:
        from .sqla.model import save_to_file
        from .protxml import getprotxml

        logger.info("consolidating eics")
        eic = consolidate_eic(envelopes)
        eic = eic.set_index("peptide_index")

        me = IO(pepdf).read_df()
        prot = getprotxml(protdf)
        n = len(me)
        if "pepxml_index" in me.columns:  # in save subset
            me = me.set_index("pepxml_index")
        me = me.join(eic, how="inner", rsuffix="_env")
        me = me.reset_index(drop=True)
        if "peptide_env" in me.columns:
            assert (me["peptide"] == me["peptide_env"]).all()
            me.drop(columns=["peptide_env"], inplace=True)
        # create explicit index for parquet
        me.index.name = "result_index"
        me.reset_index(inplace=True)  # add result_index to columns

        logger.info(
            "writing results: %s[%d] original length=%d",
            result_name.name,
            len(me),
            n,
        )
        me = dedup_peptides(me)
        save_to_file(result_name, me, prot)


# def dedup_peptides1(pep: pd.DataFrame) -> pd.DataFrame:
#     return (
#         pep.groupby(by=["peptide", "modcol"])
#         .apply(lambda x: x.sort_values(by="heavyCor", ascending=False).head(1))
#         .reset_index(drop=True)
#     )


def dedup_peptides(pep: pd.DataFrame) -> pd.DataFrame:
    """Remove duplicate peptide/modcol rows, taking smallest nnls_residual."""
    return (
        pep.sort_values(by="nnls_residual", ascending=True)
        .groupby(by=["peptide", "modcol", "assumed_charge"])
        .head(1)
        .reset_index()
    )


def consolidate_eic(eicfiles: list[Path]) -> pd.DataFrame:
    df = pd.concat(
        [IO(eic).read_df() for eic in eicfiles],
        axis="index",
        ignore_index=True,
    )
    return df


class MzMLTask(Task):
    @override
    def task_run(self) -> pd.DataFrame:
        mzmlc = MZI(self.mzml)
        done = mzml_calc_mzml_envelopes(
            self.pepxml_df,
            mzmlc,
            self.settings,
            level=self.level,
            number_of_bg_processes=self.number_of_bg_processes,
        )

        outdf = pd.DataFrame(done)
        for col in ["relativeIsotopeAbundance", "enrichment", "heavyCor"]:
            if col in outdf.columns:
                outdf[col] = outdf[col].astype(np.float32)
        return outdf


def create_mz_tasks(
    pepxml_df: pd.DataFrame,
    job: TurnoverJob,
    nspectra: int | None = None,  # spectra chunk sizes for tasks
    *,
    testing: int | None = None,
) -> list[MzMLTask]:
    tasks = []
    mz2run = job.get_mzfile_to_run()
    files = job.to_resource_files()
    todo: list[tuple[pd.DataFrame, MzMLResourceFile]] = []
    dfs: list[tuple[pd.DataFrame, MzMLResourceFile]]

    if testing is not None:
        logger.warning("TESTING has been set to ON! Output files are VOID!")
        pepxml_df = pepxml_df.iloc[:testing]
        nspectra = max(testing // 3, 2)

    for mzml in files.mzmlfiles:
        if job.match_runNames:
            df = pepxml_df[pepxml_df["run"] == mz2run[mzml.name]]
        else:
            df = pepxml_df

        if len(df) == 0:  # pragma: no cover
            continue
        # Don't need "peptide" really... just for sanity check
        df = df[
            [
                "retention_time_sec",
                "peptide",
                "maxIso",
                "mzranges",
                "isLabelledRun",
            ]
        ]
        if nspectra is not None and nspectra > 0:
            chunk_num = max(len(df) // nspectra, 1)
            if chunk_num > 1:
                dfs = [(sdf, mzml) for sdf in array_split(df, chunk_num)]
            else:
                dfs = [(df, mzml)]
        else:
            dfs = [(df, mzml)]
        todo.extend(dfs)

    ntotal = len(todo)
    tasks = [
        MzMLTask(cdf, mzml, job.settings, level, ntotal)
        for level, (cdf, mzml) in enumerate(todo, start=1)
    ]

    return tasks


def calc_rtranges(
    rundf: pd.DataFrame,
    mzmlc: MZI,
    settings: PeptideSettings,
    rt_min: int,
) -> pd.DataFrame:
    rt = mzmlc.df["retention_time_sec"]
    if mzmlc.dino is not None:  # pragma: no cover
        rundf = rundf.join(findrt_dino(mzmlc.dino, rundf, rt, settings))
        rrt = rundf.rtmax - rundf.rtmin
        logger.info("dinosaur: rt range %s-%s secs", str(rrt.min()), str(rrt.max()))
    else:
        rundf = rundf.join(findrt(rundf, rt, settings))

    search = mzmlc.search

    def inrange(mzr) -> np.bool:
        return np.any(mzr[:, 0] <= search.mzmax) and np.any(mzr[:, 1] >= search.mzmin)

    q = (
        (rundf["rtmin"] <= search.rtmax)
        & (rundf["rtmax"] >= search.rtmin)
        & (rundf["rt_count"] >= rt_min)
        & rundf["mzranges"].apply(inrange)  # type: ignore
    )

    return rundf[q]


def compute_rt(rundf: pd.DataFrame, settings: PeptideSettings):
    # if 'rtmin' in rundf.columns or 'rtmax' in rundf.columns:
    #     return rundf
    rttol = settings.rtTolerance
    rt = rundf["retention_time_sec"].apply(
        lambda rt: pd.Series(dict(rtmin=rt - rttol, rtmax=rt + rttol)),
    )
    return rundf.join(rt)


def findrt(
    rundf: pd.DataFrame,
    rt: pd.Series,
    settings: PeptideSettings,
) -> pd.DataFrame:
    use_in_sample = settings.retentionTimeCorrection == "UseInSample"
    rttol = settings.rtTolerance

    def rtrange(row: pd.Series) -> pd.Series:
        rts = row.retention_time_sec  # from pep.xml file
        if use_in_sample and row.isLabelledRun:
            rtmin = rts - rttol / 2
            rtmax = rts + rttol / 2

        else:
            rtmin = rts - rttol
            rtmax = rts + rttol

        rt_count = ((rt >= rtmin) & (rt <= rtmax)).sum()

        return pd.Series(
            dict(rtmin=rtmin, rtmax=rtmax, rt_count=float(rt_count)),
            dtype=np.float32,
        )

    return rundf[["retention_time_sec", "isLabelledRun"]].apply(rtrange, axis=1)


def findrt_dino(  # pragma: no cover
    dino_df: pd.DataFrame,
    rundf: pd.DataFrame,
    rt: pd.Series,
    settings: PeptideSettings,
) -> pd.DataFrame:
    use_in_sample = settings.retentionTimeCorrection == "UseInSample"

    dinomz = dino_df["mz"]
    rttol = settings.rtTolerance

    def findrtrange(row: pd.Series) -> pd.Series:
        total = 0
        rtmin = 1e10
        rtmax = -1.0
        rt_count = 0
        for mzin, mzmax in row.mzranges:
            q = (dinomz >= mzin) & (dinomz <= mzmax)
            n = q.sum()
            if n > 0:
                df = dino_df[q]
                rtmin = min(rtmin, df.rtStart.min())
                rtmax = max(rtmax, df.rtEnd.max())
            total += n
            if not total:
                if use_in_sample and row.isLabelledRun:
                    rtmin = row.retention_time_sec - rttol / 2
                    rtmax = row.retention_time_sec + rttol / 2
                else:
                    rtmin = row.retention_time_sec - rttol
                    rtmax = row.retention_time_sec + rttol
            rt_count = ((rt >= rtmin) & (rt <= rtmax)).sum()
        return pd.Series(
            dict(rtmin=rtmin, rtmax=rtmax, rt_count=float(rt_count)),
            dtype=np.float32,
        )

    return rundf[["mzranges", "retention_time_sec", "isLabelledRun"]].apply(
        findrtrange,
        axis=1,
    )


def mzml_calc_mzml_envelopes(
    rundf: pd.DataFrame,
    mzmlc: MZI,
    settings: PeptideSettings,
    level: int = 0,
    number_of_bg_processes: int = 1,
) -> list[dict[str, Any]]:
    from .logger import log_iterator
    from .config import MIN_RT

    n = len(rundf)
    rundf = calc_rtranges(
        rundf,
        mzmlc,
        settings,
        rt_min=MIN_RT,
    )
    logger.info(
        "%s[%d/%d]: removed out of range rt,mz %d/%d",
        mzmlc.name,
        level,
        number_of_bg_processes,
        len(rundf),
        n,
    )
    total = len(rundf)
    it = log_iterator(
        rundf.itertuples(index=True),
        total=total,
        desc=mzmlc.name,
        level=level,
        number_of_bg_processes=number_of_bg_processes,
    )
    row: PepXMLRunRowRT
    done = []
    # nhits = 0
    for row in it:  # type: ignore
        idx = row[0]
        d, fail = mzml_calc_pep_envelopes(mzmlc, row, settings)
        if d is None:
            continue
        d["peptide_index"] = idx
        d["peptide"] = row.peptide  # don't need this just sanity check
        d["fail"] = fail
        # d['runName'] = mzmlc.mzmlfile.runName

        done.append(d)
    logger.info(
        "%s[%d/%d]: found %d/%d envelopes",
        mzmlc.name,
        level,
        number_of_bg_processes,
        len(done),
        total,
    )
    return done


def mzml_calc_pep_envelopes(
    mzmlc: MZI,
    row: PepXMLRunRowRT,
    settings: PeptideSettings,
) -> tuple[dict[str, Any] | None, int]:
    from .config import MIN_RT

    rtmin, rtmax = row.rtmin, row.rtmax
    debug = logger.isEnabledFor(logging.DEBUG)

    start = datetime.now()
    eic = mzmlc.getEIC(row.mzranges, rtmin, rtmax, minrt=MIN_RT)
    if eic is None:  # pragma: no cover
        return None, 1 << FAIL_E_NOEICS
    end = datetime.now()

    if debug:  # pragma: no cover
        logger.debug("getEIC: %s %s %d", row.peptide, end - start, len(eic))

    maxIso = row.maxIso
    assert len(eic) == maxIso + 2

    start = end
    e, fail = fitEnvelope(row.peptide, eic, settings)
    if e is None:
        return None, fail
    if debug:  # pragma: no cover
        logger.debug("fitEnvelope: %s %s", row.peptide, datetime.now() - start)
    d = dict(eics=eic)
    d.update(asdict(e))
    return d, fail


def turnover_prepare(
    files: ResourceFiles,
    force: bool = False,
    workers: int = 1,
) -> None:
    """prepare pepXML and mzmlFiles"""
    from functools import partial
    from .pepxml import pepxml_create
    from .protxml import protxml_create
    from .parallel_utils import parallel_result

    if len(files.mzmlfiles) == 0:  # pragma: no cover
        logger.warning("no mzML files!")
        return

    files.ensure_directories()

    procs: list[
        tuple[
            Callable[[Any, int, int], int],
            MzMLResourceFile | PepXMLResourceFile | ProtXMLResourceFile,
        ]
    ] = []

    for pepxml in files.pepxmls:
        if force or not pepxml.cache_pepxml_ok():
            procs.append((pepxml_create, pepxml))

        else:
            logger.info(
                'skipping creation of "%s" cache: %s',
                pepxml.name,
                pepxml.cache_pepxml().name,
            )
    if files.protxml.exists():
        if force or not files.protxml.cache_protxml_ok():
            procs.append((protxml_create, files.protxml))

        else:
            logger.info(
                'skipping creation of "%s" cache: %s',
                files.protxml.name,
                files.protxml.cache_protxml().name,
            )
    todo = (
        [t for t in files.mzmlfiles if not t.cache_mzml_ok()]
        if not force
        else files.mzmlfiles
    )
    skipped = set(files.mzmlfiles) - set(todo)
    if skipped:  # pragma: no cover
        for t in skipped:
            logger.info(
                'skipping creation of "%s" cache: %s',
                t.name,
                t.cache_mzml().name,
            )
    target: MzMLResourceFile | PepXMLResourceFile | ProtXMLResourceFile
    for target in todo:
        procs.append((mzml_create, target))

    if not procs:  # pragma: no cover
        return
    cleanups: dict[int, BaseResourceFile] = {}
    exe: list[Callable[[], int]] = []
    for idx, (func, target) in enumerate(procs, start=1):
        exe.append(partial(func, target, idx, len(procs)))
        cleanups[idx] = target

    ntotal = len(exe)
    try:
        for ridx in parallel_result(exe, workers=workers):
            logger.info("turnover_prepare task done: [%d/%d]", ridx, ntotal)
            if ridx in cleanups:
                del cleanups[ridx]

    except KeyboardInterrupt:  # pragma: no cover
        for c in cleanups.values():
            c.cleanup()
        raise



from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass

import numpy as np
import pandas as pd
from typing_extensions import override


def and_filters(filters: list[pd.Series[bool]]) -> pd.Series[bool]:
    if not filters:
        raise ValueError("no filters")
    if len(filters) == 1:
        return filters[0]
    return np.bitwise_and.reduce(filters)


class BaseFilter(ABC):
    @classmethod
    @abstractmethod
    def columns(cls) -> list[str]:  # pragma: no cover
        return []

    def filter(self, df: pd.DataFrame, *, copy: bool = False) -> pd.DataFrame:
        b = self.booleans(df)
        if b is None:  # pragma: no cover
            return df
        return df[b].copy() if copy else df[b]

    @abstractmethod
    def booleans(self, df: pd.DataFrame) -> pd.Series | None:  # pragma: no cover
        return None

    def and_filters(self, filters: list[pd.Series[bool]]) -> pd.Series[bool]:
        return and_filters(filters)


@dataclass
class PepXMLFilter(BaseFilter):
    noX: bool = True
    minProbabilityCutoff: float | None = 0.8

    @override
    @classmethod
    def columns(cls) -> list[str]:  # pragma: no cover
        return ["peptide", "peptideprophet_probability", "probability"]

    @override
    def booleans(self, df: pd.DataFrame) -> pd.Series | None:
        filters: list[pd.Series[bool]] = []

        def add(b: pd.Series[bool]) -> None:
            filters.append(b)

        if self.noX:
            add(~df["peptide"].str.contains("X"))

        if self.minProbabilityCutoff is not None:
            if "peptideprophet_probability" in df:
                add(df["peptideprophet_probability"] >= self.minProbabilityCutoff)
            elif "probability" in df:  # pragma: no cover
                add(df["probability"] >= self.minProbabilityCutoff)
        if not filters:  # pragma: no cover
            return None
        return pd.Series(self.and_filters(filters), index=df.index)



from __future__ import annotations

import logging
import mmap
import re
from collections import Counter
from pathlib import Path
from typing import Callable
from typing import Iterator

import numpy as np
import pandas as pd

from .algo import settings_to_algo
from .jobs import TurnoverJob
from .logger import logger
from .resourcefiles import PepXMLResourceFile
from .types.pepxmltypes import PepXMLRunRow
from .utils import Apply
from .utils import human
from .utils import IO
from .utils import PeptideSettings

# CALL TREE


# get_and_filter_pepxml
# | --- getpepxml
#            | ---- pepxml_create (if no cache)
#                        | --- pepxml_raw # pep.xml to df
#                        | --- pepxml_out # df -> parquet
#                                 | --- IO.write_df
#            | IO.read_df (from cache)
#  | --- filter_pepxml
#         filter X
#              | --- compute_mz_modcol
#                          | --- compute_mz
#
#              | -- compute_maxIso (and mzranges)
#  | --- aggregate_search_hits
#
#
def bchop_spectrum(spectrum: bytes) -> bytes:
    ret = b".".join(spectrum.split(b".")[:-3])
    return ret


SPECTRUM = re.compile(b'spectrum="([^"]+)"')


## @export
def count_spectra(pepxml: Path) -> dict[str, int]:
    """Very fast inspection of pep.xml spectra names"""
    cnt: dict[bytes, int] = Counter()
    with pepxml.open("rb") as fp:
        with mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            for m in SPECTRUM.finditer(mm):
                spectrum = bchop_spectrum(m.group(1))
                cnt[spectrum] += 1

    return {k.decode("ascii"): v for k, v in cnt.items()}


SPECTRUM_QUERY = re.compile(b"<spectrum_query ")


def scan_spectra(pepxml: Path) -> Iterator[int]:
    """Very fast inspection of pep.xml spectra queries"""
    with pepxml.open("rb") as fp:
        with mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            for m in SPECTRUM_QUERY.finditer(mm):
                yield m.start(0)


PEPTIDE_PROPHET_QUERY = re.compile(b"<peptideprophet_summary ")


## @export
def scan_pp_probability(pepxml: Path) -> bool:
    with pepxml.open("rb") as fp:
        with mmap.mmap(fp.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            for _ in PEPTIDE_PROPHET_QUERY.finditer(mm):
                return True
    return False  # pragma: no cover


def decoy_prefix_fn(prefix: str = "DECOY_") -> Callable[[list[str]], bool]:
    prefix = prefix.lower()
    return lambda proteins: all(p.lower().startswith(prefix) for p in proteins)


def decoy_postfix_fn(postfix: str = "_DECOY") -> Callable[[list[str]], bool]:
    postfix = postfix.lower()
    return lambda proteins: all(p.lower().endswith(postfix) for p in proteins)


def compute_mz(df: pd.DataFrame) -> pd.DataFrame:
    # add H+
    HPLUS = 1.00727646677
    df["observed_mz"] = df["precursor_neutral_mass"] / df["assumed_charge"] + HPLUS

    df["mz"] = df["calc_neutral_pep_mass"] / df["assumed_charge"] + HPLUS
    return df


def compute_labelledRuns(df: pd.DataFrame, runNames: set[str] | None = None):
    if runNames is not None:
        df["isLabelledRun"] = df["run"].apply(lambda r: r in runNames)
    else:
        df["isLabelledRun"] = False
    return df


def compute_modcol(
    df: pd.DataFrame,
    decoy_prefix: str = "DECOY_",
    decoy_postfix: str | None = None,
) -> pd.DataFrame:
    from .utils import calculate_fdrs

    decoyfn = (
        decoy_postfix_fn(decoy_postfix)
        if decoy_postfix is not None
        else decoy_prefix_fn(decoy_prefix)
    )
    df["modcol"] = df["modifications"].apply(Apply.modcol)
    if "is_decoy" not in df.columns:
        df["is_decoy"] = df["proteins"].apply(decoyfn)
    if "fdr" not in df.columns:
        for prob in ["peptideprophet_probability", "probability"]:
            if prob in df.columns:
                df["fdr"] = calculate_fdrs(df, score=prob)
                break

    return df


def unique_seq(some_list: list[list[str]]) -> list[str]:
    return list({s for lv in some_list for s in lv})


def strlist_to_str(some_list: list[str]) -> str:
    return ":".join(sorted(set(some_list)))


# cols = ['spectrum', 'spectrumNativeID', 'precursor_neutral_mass',
#    'assumed_charge', 'retention_time_sec', 'start_scan', 'end_scan',
#    'index', 'proteins', 'protein_descr', 'peptide_next_aa',
#    'peptide_prev_aa', 'num_tol_term', 'xcorr', 'deltacn', 'deltacnstar',
#    'spscore', 'sprank', 'expect', 'modifications', 'hit_rank', 'peptide',
#    'num_tot_proteins', 'num_matched_ions', 'tot_num_ions',
#    'num_missed_cleavages', 'calc_neutral_pep_mass', 'massdiff',
#    'num_matched_peptides', 'modified_peptide', 'fval', 'ntt', 'nmc',
#    'massd', 'isomassd', 'peptideprophet_probability',
#    'peptideprophet_ntt_prob', 'is_rejected']
AGG = dict(
    calc_neutral_pep_mass="median",
    precursor_neutral_mass="median",
    mz="median",
    observed_mz="median",
    retention_time_sec="median",
    num_missed_cleavages="max",
    peptideprophet_probability="max",
    interprophet_probability="max",
    labelledElementCount="max",
    is_decoy="any",
    agg_count="sum",
    proteins=unique_seq,
    protein_descr=unique_seq,
    modifications="first",
    xcorr="max",  # COMET searchEngineScore
    ionscore="max",  # MASCOT searchEngineScore
    run="first",  # strlist_to_str,
    isLabelledRun="any",
    searchEngineScore="max",
    fdr="max",
    # PROTXML ....
    probability="max",
    group_number="nunique",
    percent_coverage="max",
    confidence="max",
    protein_description=unique_seq,
    ngroups="max",
    hit_rank="min",
    num_matched_ions="max",
    tot_num_ions="max",
    has_real="any",
    isomassd="max",
    modified_peptide="first",
)


def aggregate_search_hits(
    df: pd.DataFrame,
    *,
    settings: PeptideSettings,
) -> pd.DataFrame:
    mztol = settings.mzTolerance

    # use_simple_median = settings.retentionTimeCorrection != "UseInSample"

    df["agg_count"] = 1

    assert "modcol" in df.columns

    keys = ["peptide", "_massint", "assumed_charge", "modcol"]
    df["_massint"] = df["calc_neutral_pep_mass"].apply(lambda m: round(m / (2 * mztol)))
    remove = ["_massint"]

    # only values that exist
    agg = {col: v for col, v in AGG.items() if col in df.columns and col not in keys}
    missing = [col for col in df.columns if col not in agg and col not in keys]
    if missing:
        logger.warning('aggregate columns set to choose "first": %s', missing)
        for col in missing:
            agg[col] = "first"
    cols = keys + list(agg)

    g = df[cols].groupby(by=keys, as_index=False)

    ret_df = g.aggregate(agg)

    ret_df = ret_df.drop(columns=remove)
    return ret_df


def pepxml_raw(
    pepxml: PepXMLResourceFile,
    level: int = 0,
    number_of_bg_processes: int = 1,
) -> pd.DataFrame:
    # from .broken_api import PepXMLDataFrame
    from .pepxml_reader import pepxml_dataframe

    logger.info("reading: %s", pepxml.original.name)
    df = pepxml_dataframe(
        pepxml.original,
        level=level,
        number_of_bg_processes=number_of_bg_processes,
    )

    logger.info("done: %s", pepxml.original.name)
    return df


def compute_maxIso(
    df: pd.DataFrame,
    settings: PeptideSettings,
) -> pd.DataFrame:
    def element_count(peptide: str) -> int:
        return settings.getElementCountFromPeptide(peptide)

    # max_iso = mk_maxIso(settings)
    max_iso = settings_to_algo(settings).mk_maxIso(settings)

    def mzranges(pep: PepXMLRunRow) -> np.ndarray:
        return settings.eic_mzranges(pep)

    logger.info(
        "compute_maxIso: adding labelledElementCount, maxIso, mzranges",
    )
    df["labelledElementCount"] = df["peptide"].apply(element_count)
    # get rid of peptides with no labelled Elements
    zero = df["labelledElementCount"] == 0
    anyz = zero.sum()
    if anyz > 0:
        logger.warning("removing %s peptides with no labelled element", anyz)
        df = df[~zero].copy()
    df["maxIso"] = df["peptide"].apply(max_iso)
    # ndarray[maxIso+2,2] of mzmin,mzmax
    df["mzranges"] = df[["maxIso", "assumed_charge", "mz", "observed_mz"]].apply(
        mzranges,
        axis=1,
    )
    logger.info("compute_maxIso: done")
    return df


def getpepxml(
    pepxml: PepXMLResourceFile,
) -> pd.DataFrame:
    if not pepxml.cache_pepxml_ok():
        if not pepxml.exists():
            raise FileNotFoundError(str(pepxml.original))
        logger.info('getpepxml: creating cache "%s"', pepxml.original.name)
        pepxml_create(pepxml)
    assert pepxml.cache_pepxml_ok()
    data = pepxml.cache_pepxml()
    logger.info('getpepxml: reading: "%s"', data.name)
    df = IO(data).read_df()
    return df


def filter_pepxml(df: pd.DataFrame, job: TurnoverJob) -> pd.DataFrame:
    from .filters import PepXMLFilter

    if job.match_runNames:  # pragma: no cover
        n = len(df)
        df = df[df["run"].isin(job.runNames)].copy()
        logger.info('filter_pepxml: run names: "%s" [%d/%d]', job.runNames, len(df), n)

    filter = PepXMLFilter(minProbabilityCutoff=job.settings.minProbabilityCutoff)
    df = filter.filter(df, copy=True)

    return df


def compute_pre_pepxml(pepxml_df: pd.DataFrame, job: TurnoverJob) -> pd.DataFrame:
    pepxml_df = compute_labelledRuns(
        pepxml_df,
        runNames=job.runNames if job.match_runNames else None,
    )
    # needed for grouping
    pepxml_df = compute_modcol(pepxml_df)
    return pepxml_df


def compute_post_pepxml(pepxml_df: pd.DataFrame, job: TurnoverJob) -> pd.DataFrame:
    pepxml_df = compute_mz(pepxml_df)  # add mz and observed_mz

    pepxml_df = compute_maxIso(pepxml_df, job.settings)
    return pepxml_df


def get_and_filter_pepxml(job: TurnoverJob) -> pd.DataFrame:
    files = job.to_resource_files()
    ret = []
    for pepxml in files.pepxmls:
        pepxml_df = getpepxml(pepxml)

        pepxml_df = filter_pepxml(pepxml_df, job)
        ret.append(pepxml_df)
    pepxml_df = pd.concat(ret, axis=0, ignore_index=True)

    pepxml_df = compute_pre_pepxml(pepxml_df, job)
    if job.aggregate_peptides and "agg_count" not in pepxml_df:
        logger.info("aggregating: initial number=%d", len(pepxml_df))
        pepxml_df = aggregate_search_hits(
            pepxml_df,
            settings=job.settings,
        )
        logger.info("aggregating done: final number=%d", len(pepxml_df))
    # compute extra columns after any aggregation
    return compute_post_pepxml(pepxml_df, job)


def pepxml_out(
    df: pd.DataFrame,
    pepxml: PepXMLResourceFile,
) -> None:
    from .utils import IO

    data = pepxml.cache_pepxml()

    if logger.isEnabledFor(logging.INFO):  # pragma: no cover
        if "agg_count" in df.columns:
            logger.info(  # pylint: disable=logging-fstring-interpolation
                f"writing file: {data.name}: total={len(df)} multiple={(df.agg_count > 1).sum()}",
            )
        else:
            logger.info("writing file: %s: total=%s", data.name, len(df))

    # df = dehydrate_pepxml(df)
    assert "mzranges" not in df.columns
    IO(data, df).save_df()

    if logger.isEnabledFor(logging.INFO):  # pragma: no cover
        mem = df.memory_usage(deep=True)
        size = data.stat().st_size
        osize = pepxml.original.stat().st_size
        runs = ",".join(sorted(df.run.unique()))
        logger.info(  # pylint: disable=logging-fstring-interpolation
            f"memory={human(mem.sum())} disk={human(size)} original={human(osize)}: runs={runs}",
        )


def pepxml_create(
    pepxml: PepXMLResourceFile,
    level: int = 0,
    number_of_bg_processes: int = 1,
) -> int:
    df = pepxml_raw(
        pepxml,
        level=level,
        number_of_bg_processes=number_of_bg_processes,
    )
    pepxml_out(df, pepxml)
    return level


def dehydrate_pepxml(df: pd.DataFrame) -> pd.DataFrame:
    if "mzranges" in df.columns:
        df["mzranges"] = df["mzranges"].apply(lambda x: x.flatten())
    return df



from __future__ import annotations

import os
import signal
import subprocess
import sys
import time
import tomllib
from datetime import datetime
from pathlib import Path
from shutil import rmtree
from shutil import which
from typing import Any
from typing import Iterator
from typing import Sequence

import tomli_w
from typing_extensions import override

from .logger import logger

try:
    from psutil import pid_exists  # type: ignore

    psutil_ok = True  # pragma: no cover
except ImportError:
    psutil_ok = False

    # for windows we need psutil.pid_exists(pid)
    def pid_exists(pid: int) -> bool:  # pragma: no cover
        """Check whether pid exists in the current process table Unix Only."""
        if pid == 0:
            # According to "man 2 kill" PID 0 has a special meaning:
            # it refers to <<every process in the process group of the
            # calling process>> so we don't want to go any further.
            # If we get here it means this UNIX platform *does* have
            # a process with id 0.
            return True
        try:
            os.kill(pid, 0)
        except ProcessLookupError:
            return False
        except PermissionError:
            # EPERM clearly means there's a process to deny access to
            return True
        # According to "man 2 kill" possible error values are
        # (EINVAL, EPERM, ESRCH)
        else:
            return True


# see https://stackoverflow.com/questions/35772001/how-to-handle-a-signal-sigint-on-a-windows-os-machine


# from https://stackoverflow.com/questions/18499497/how-to-process-sigterm-signal-gracefully
# class GracefulKiller:
#     kill_now: bool = False

#     def __init__(self) -> None:
#         self.kill_now: bool = False
#         self.old: Any = None

#     def __enter__(self) -> None:
#         self.old = signal.signal(signal.SIGTERM, self.exit_gracefully)

#     def __exit__(self, *args: Any) -> None:
#         if self.old is not None:
#             signal.signal(signal.SIGTERM, self.old)
#             self.old = None
#         if self.kill_now:
#             raise ContinueException()

#     def exit_gracefully(self, *args: Any) -> None:
#         self.kill_now = True


class ContinueException(Exception):
    pass


class Handler:
    def __init__(self, processing: bool = True):
        self.processing = processing

    def __call__(self, signum: int, frame: Any) -> None:  # pragma: no cover
        if self.processing:
            return
        raise ContinueException()

    def arm(self) -> None:  # pragma: no cover
        if sys.platform == "win32":
            return
        signal.signal(signal.SIGCONT, self)


PID_FILENAME = "turnover.pid"
PROCESSING = "turnover.pid.is_running"


class Processing:
    """Is the background worker processing a job ATM?"""

    def __init__(self, jobsdir: Path):
        self._processing = jobsdir.joinpath(PROCESSING)

    def is_processing(self) -> bool:
        return self._processing.exists()

    def set_processing(self, start: bool) -> None:
        if start:
            self._processing.touch()
        else:
            try:
                self._processing.unlink(missing_ok=True)
            except OSError:  # pragma: no cover
                pass


## @export
class SimpleQueueClient:
    """Used by web client to interact with background Queue.

    Currently only just sends a signal to wake up.
    """

    def __init__(self, jobdir: Path) -> None:
        self.jobdir = jobdir
        self.pidfile = jobdir.joinpath(PID_FILENAME)

    def terminate(self) -> bool:  # pragma: no cover
        pid = self.get_pid()
        if pid is None:
            return False
        try:
            os.kill(pid, signal.SIGINT)
        except ProcessLookupError:
            return False
        except PermissionError:
            return False
        return True

    def signal(self) -> bool:  # pragma: no cover
        pid = self.get_pid()
        if pid is None:
            return False
        self.pidfile.touch()
        if sys.platform == "win32":
            return False
        try:
            os.kill(pid, signal.SIGCONT)
        except ProcessLookupError:
            return False
        except PermissionError:
            return False
        return True

    def is_running(self) -> bool:
        pid = self.get_pid()
        if pid is None:
            return False
        return self._pid_exists(pid)

    def get_pid(self) -> int | None:
        if self.pidfile.exists():
            with self.pidfile.open("r") as fp:
                try:
                    return int(fp.read())
                except (TypeError, OSError):  # pragma: no cover
                    return None
        return None  # pragma: no cover

    def _pid_exists(self, pid: int) -> bool:
        return pid_exists(pid)


def remove_old_sqlite(root_list: list[Path], seconds: float) -> int:
    """Remove any "old" .sqlite files but only if there is a .gz file
    next to it. This is used to save space when --compress-result is used."""
    from .exts import RESULT_EXT

    now = datetime.now().timestamp()
    n = 0
    for root in root_list:
        for d, _, files in root.walk():
            for f in files:
                if f.endswith(RESULT_EXT):
                    sqlite = d / f
                    sqlitegz = sqlite.with_suffix(sqlite.suffix + ".gz")
                    # if this is a symlink then *assume* the target is there
                    # on network filesystem
                    if not sqlitegz.exists(follow_symlinks=False):
                        continue
                    if not sqlite.exists(
                        follow_symlinks=True,
                    ):  # sqlite might be symlink
                        continue
                    delta = now - sqlite.stat(follow_symlinks=True).st_mtime
                    # older than seconds
                    if delta > seconds:
                        try:
                            # remove underlying file
                            rf = sqlite.resolve()
                            rf.unlink(missing_ok=True)
                            logger.info('removed old sqlite file "%s"', rf)
                            n += 1
                        except OSError:
                            pass
    return n


class SimpleQueue:
    INTERROR = 22

    def __init__(
        self,
        jobdirs: Sequence[str | Path] | str | Path,
        wait: float = 60,
        workers: int = 4,
        nice: int = 0,
        config: str | None = None,
        mailhost: str | None = None,
        compress_result: bool = False,
    ) -> None:
        if isinstance(jobdirs, (str, Path)):
            jobdirs = [jobdirs]
        assert len(jobdirs) > 0, "must have at least one job directory"
        self.jobsdirs = list(Path(d) for d in jobdirs)

        self.jobdir = self.jobsdirs[0]
        self.wait = wait
        self.workers = workers
        self.nice = nice
        self.mailhost = mailhost
        self.compress_result = compress_result
        self.nice_cmd = which("nice")
        self.handler = Handler()
        if config is not None:
            if not os.path.exists(config):  # pragma: no cover
                logger.warning(
                    "configuration file %s doesn't exist! ignoring...",
                    config,
                )
                config = None
        self.config = config
        self.processing = Processing(self.jobdir)

        self.quit = False  # only for testing
        self.arm_signals()

    def arm_signals(self) -> None:  # pragma: no cover
        self.handler.arm()

    def start_cleanup(self) -> None:
        """scan jobdir and remove any .sqlite files older than age seconds
        but only if there is a .gz file next to it. This runs in a separate thread.
        """
        from threading import Thread
        from .config import COMPRESS_AGE, COMPRESS_SLEEP

        def cleanup() -> None:
            while True:
                remove_old_sqlite(self.jobsdirs, seconds=COMPRESS_AGE)
                time.sleep(COMPRESS_SLEEP)

        tr = Thread(target=cleanup, daemon=True)
        tr.start()

    def command(self, *args: str) -> list[str]:  # pragma: no cover
        nice = []
        if self.nice_cmd and self.nice:
            nice = [self.nice_cmd, "-n", str(self.nice)]

        config = [f"--config={self.config}"] if self.config is not None else []

        email = [f"--mailhost={self.mailhost}"] if self.mailhost else []
        if self.compress_result:
            email.append("--compress-result")

        return [
            *nice,
            sys.executable,
            "-m",
            "protein_turnover",
            *config,
            "run",
            f"--interrupt-as-error={self.INTERROR}",
            f"--workers={self.workers}",
            *email,
            *args,
        ]

    def pidfile(self, tomlfile: Path) -> Path:
        return tomlfile.parent.joinpath(tomlfile.name + ".pid")

    def rmlog(self, tomlfile: Path) -> None:
        lf = tomlfile.parent.joinpath(tomlfile.stem + ".log")
        lf.unlink(missing_ok=True)

    def runjobs(self, it: Iterator[Path]) -> None:
        for tomlfile in it:
            self.rmlog(tomlfile)
            cmd = self.command(str(tomlfile))
            with subprocess.Popen(
                cmd,
                shell=False,
                text=True,
            ) as proc:
                try:
                    # logger.info("%s: running pid=%d", tomlfile, proc.pid)
                    logger.info("running[%s]: %s", proc.pid, cmd)
                    # let website know that this job is running....
                    # see SimpleQueueClient
                    pid = self.pidfile(tomlfile)
                    with pid.open("w") as fp:
                        fp.write(str(proc.pid))
                    self.processing.set_processing(True)
                    try:
                        _, errs = proc.communicate()
                        if errs:
                            logger.error("error from %s: %s", tomlfile, errs)
                        ret = proc.wait()  # TODO what happens when we're terminated?
                    finally:  # can catch KeyboardInterrupt
                        pid.unlink(missing_ok=True)
                        self.processing.set_processing(False)
                    if ret < 0:  # pragma: no cover
                        if -ret in {signal.SIGTERM, signal.SIGKILL}:
                            logger.warning("%s: killed...", tomlfile)
                    status = (
                        "finished"
                        if ret == 0
                        else ("killed" if ret < 0 or ret == self.INTERROR else "failed")
                    )
                    logger.info("%s: status=%s", tomlfile, status)
                    try:
                        self.update_status(tomlfile, status)
                    except Exception as e:  # pragma: no cover
                        logger.error("can't update status! %s: %s", tomlfile, e)
                except KeyboardInterrupt:  # pragma: no cover
                    logger.info("sending signal to child process")
                    if sys.platform == "win32":
                        proc.terminate()
                        rmtree(tomlfile.parent, ignore_errors=True)
                    else:
                        proc.send_signal(signal.SIGINT)
                    try:
                        proc.wait(1.0)
                    except subprocess.TimeoutExpired:
                        pass

    def update_status(self, tomlfile: Path, status: str) -> None:
        c = self.read_toml(tomlfile)
        if c is None:  # pragma: no cover
            return
        c["status"] = status
        if status in {"finished", "killed", "failed"}:
            c["finished_time"] = datetime.now()
        with tomlfile.open("wb") as fp:
            tomli_w.dump(c, fp)

    def read_toml(self, tomlfile: Path) -> dict[str, Any] | None:
        try:
            with tomlfile.open("rb") as fp:
                return tomllib.load(fp)
        except Exception as e:  # pragma: no cover
            logger.error("can't open %s: %s", tomlfile, e)
            return None

    def status(self, tomlfile: Path) -> str:
        c = self.read_toml(tomlfile)
        if c is None:  # pragma: no cover
            return "failed"
        return str(c.get("status", "stopped"))

    def search(self, directory: Path, wait: float = 60.0) -> Iterator[Path]:
        nloop = 0
        mtime = None
        todo: list[tuple[Path, float]] = []
        every = max(60 * int(60 / wait), 1) if wait else 1  # each hour
        while True:
            if (nloop % every) == 0:
                logger.info("searching for jobs in %s", directory.absolute())
            nloop += 1
            last = self.mtime()
            for d, _, files in directory.walk():
                for f in files:
                    if f.endswith(".toml"):
                        tomlfile = d / f
                        # already running
                        if self.pidfile(tomlfile).exists():
                            continue
                        mod = tomlfile.stat().st_mtime
                        if mtime is None or mod > mtime:
                            try:
                                if self.status(tomlfile) == "pending":
                                    todo.append((tomlfile, mod))
                            except Exception as e:  # pragma: no cover
                                logger.error("%s: %s", tomlfile, e)

            if todo:
                n = len(todo)
                logger.info("found %d job%s", n, "" if n == 1 else "s")
                todo = sorted(todo, key=lambda t: t[1])  # oldest first
                for tomlfile, m in todo:
                    if tomlfile.exists():
                        mtime = m
                        yield tomlfile
                todo = []
            try:
                # when we are signaled by website with signal.SIGCONT
                # the self.handler will throw a ContinueException
                # but only if processing is False
                # otherwise it will eat the signal
                self.handler.processing = False
                for _ in range(20):
                    if self.quit:
                        return
                    time.sleep(wait)
                    latest = self.mtime()
                    if latest > last:
                        break
            except ContinueException:  # pragma: no cover
                # kill -CONT $(cat {pidfile})
                logger.info("awakened by signal....")
            finally:
                self.handler.processing = True

    def mtime(self):
        pidfile = self.thispid()
        return pidfile.stat().st_mtime

    def thispid(self) -> Path:
        return self.jobdir.joinpath(PID_FILENAME)

    def process(self, jobdir: Path, wait: float = 60.0) -> None:
        self.runjobs(self.search(jobdir, wait))

    def run(self) -> None:
        if self.compress_result:
            self.start_cleanup()
        pidfile = self.thispid()
        try:
            with pidfile.open("wt", encoding="utf-8") as fp:
                pid = os.getpid()
                fp.write(str(pid))
            w = "with 👍" if psutil_ok else "without 👎"
            print(f"protein_turnover queue running ({w} psutil) as pid={pid}")
            self.process(self.jobdir, self.wait)
        finally:
            pidfile.unlink(missing_ok=True)


class TestSimpleQueue(SimpleQueue):
    @override
    def arm_signals(self) -> None:
        pass

    @override
    def command(self, *args: str) -> list[str]:
        return [sys.executable, "-m", "protein_turnover.background", *args]


if __name__ == "__main__":
    # used in testing....
    import click

    @click.command()
    @click.argument("jobfile")
    def testrun(jobfile: str):
        """Minimal job run"""
        tomlfile = Path(jobfile)
        with tomlfile.open("rb") as fp:
            tf = tomllib.load(fp)

        time.sleep(0.5)  # do some computation :)
        tf["data"] = "done"
        with tomlfile.open("wb") as fp:
            tf = tomli_w.dump(tf, fp)

    testrun()



## all imports that protein_turnover_website uses
from __future__ import annotations

__all__ = [
    "TurnoverJob",
    "plot_all",
    "Column",
    "DTQuery",
    "FormFilter",
    "RowQuery",
    "RowFilter",
    "dict_to_form_filter",
    "count_spectra",
    "scan_pp_probability",
    "scan_proteins",
    "scan_mzml_spectra",
    "get_isotope_numbers",
    "get_element_names",
    "PeptideSettings",
    "SimpleQueueClient",
    "all_plotting_columns",
    "enrichment_plot",
    "nnls_plot",
    "plotLPF",
    "to_image_url",
    "file2engine",
    "Aggregate",
    "PeptideQuery",
    "ProteinQueryResult",
    "ProteinQuery",
    "RatioAggregate",
    "SimplePeptideQuery",
    "RESULT_EXT",
    "slugify",
    "calc_stats",
    "MINRSQ100",
    "sqlite_version",
    "fixsqlite",
    "remap_job",
    "decompressto",
]

from .jobs import TurnoverJob
from .jobs import slugify
from .jobs import remap_job

from .plotting import plot_all
from .plotting import enrichment_plot
from .plotting import nnls_plot
from .plotting import plotLPF
from .plotting import to_image_url

from .pepxml import count_spectra
from .pepxml import scan_pp_probability

from .protxml import scan_proteins

from .pymz import scan_mzml_spectra

from .utils import decompressto
from .utils import get_element_names
from .utils import get_isotope_numbers
from .utils import PeptideSettings

from .background import SimpleQueueClient

from .exts import RESULT_EXT


from .sqla.query import Column
from .sqla.query import DTQuery
from .sqla.query import FormFilter, all_plotting_columns
from .sqla.query import dict_to_form_filter
from .sqla.query import RowFilter
from .sqla.query import RowQuery
from .sqla.query import Aggregate
from .sqla.query import PeptideQuery
from .sqla.query import ProteinQuery
from .sqla.query import ProteinQueryResult
from .sqla.query import RatioAggregate
from .sqla.query import SimplePeptideQuery
from .sqla.query import sqlite_version
from .sqla.query import calc_stats

from .sqla.model import MINRSQ100

from .sqla.utils import file2engine

from .sqla.iso_peaks import fixsqlite
